﻿using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Ast.Visitor;
using SharQL.Algebra.AstUtils.Visitor;
using SharQL.Algebra.Tests.Utils;

namespace SharQL.Algebra.AstUtils.Tests
{
	/// <summary>
	/// Summary description for ListReificationTest
	/// </summary>
	[TestClass]
	public class ListReificationTest
	{
		public ListReificationTest()
		{
		}

		private TestContext testContextInstance;

		/// <summary>
		///Gets or sets the test context which provides
		///information about and functionality for the current test run.
		///</summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		//
		// You can use the following additional attributes as you write your tests:
		//
		// Use ClassInitialize to run code before running the first test in the class
		// [ClassInitialize()]
		// public static void MyClassInitialize(TestContext testContext) { }
		//
		// Use ClassCleanup to run code after all tests in a class have run
		// [ClassCleanup()]
		// public static void MyClassCleanup() { }
		//
		// Use TestInitialize to run code before running each test 
		// [TestInitialize()]
		// public void MyTestInitialize() { }
		//
		// Use TestCleanup to run code after each test has run
		// [TestCleanup()]
		// public void MyTestCleanup() { }
		//
		#endregion

		[TestMethod]
		public void TestListReification()
		{
			TriplesBlockNode root = createTripleWithList();
			INode oldRoot = root.Clone();
			reificateCollections(ref root);
			Visualizer.Visualize.Ast(oldRoot, root);
			verifyReification(root);
		}

		[TestMethod]
		public void TestNestedListReification()
		{
			TriplesBlockNode root = createTripleWithNestedList();
			reificateCollections(ref root);
			verifyNestedReification(root);
		}

		[TestMethod]
		public void TestEmptyListReification()
		{
			TriplesBlockNode root = createTripleWithEmptyList();
			reificateCollections(ref root);
			verifyEmptyReification(root);
		}

		private TriplesBlockNode createTripleWithNestedList()
		{
			return new TriplesBlockNode(
				new TriplesSameSubjectNode(
					new PrefixedNameNode() { Pn_Local = "x" },
					null,
					new PropertyListNode(
						new VarNode("p"),
						new ObjectListNode(
							new CollectionNode(
								new GraphNodeListNode(
									new VarNode("v"),
									new GraphNodeListNode(
										new CollectionNode(
											new GraphNodeListNode(
												new PrefixedNameNode() { Pn_Local = "a" },
												new GraphNodeListNode(
													new PrefixedNameNode() { Pn_Local = "b" },
													null
												)
											)
										),
										new GraphNodeListNode(
											new PrefixedNameNode() { Pn_Local = "w" },
											null
										)
									)
								)
							),
							null
						),
						null
					)
				),
				null
			);
		}

		private TriplesBlockNode createTripleWithEmptyList()
		{
			return new TriplesBlockNode(
				new TriplesSameSubjectNode(
					new PrefixedNameNode() { Pn_Local = "x" },
					null,
					new PropertyListNode(
						new VarNode("p"),
						new ObjectListNode(null, null),
						null
					)
				),
				null
			);
		}

		private TriplesBlockNode createTripleWithList()
		{
			return new TriplesBlockNode(
				new TriplesSameSubjectNode(
					new PrefixedNameNode() { Pn_Local = "x" },
					null,
					new PropertyListNode(
						new VarNode("p"),
						new ObjectListNode(
							new CollectionNode(
								new GraphNodeListNode(
									new VarNode("v"),
									new GraphNodeListNode(
										new PrefixedNameNode() { Pn_Local = "w" },
										null
									)
								)
							),
							null
						),
						null
					)
				),
				null
			);
		}

		private void reificateCollections(ref TriplesBlockNode root)
		{
			IVisitor visitor = new ListReificationVisitor();
			root.Accept(visitor);
		}

		private void verifyReification(TriplesBlockNode root)
		{
			TriplesBlockNode[] triples = TripleUtils.FetchSubsequentTriples(root);
			List<TriplesBlockNode> triplesList = new List<TriplesBlockNode>(triples);

			Assert.AreEqual(5, triplesList.Count);

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new PrefixedNameNode() { Pn_Local = "x" },
				new VarNode("p"),
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new VarNode("v")));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new PrefixedNameNode() { Pn_Local = "w" }));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" }));

			Assert.AreEqual(0, triplesList.Count);

			triplesList = new List<TriplesBlockNode>(triples);

			assertBlankNodesInLists(triplesList);
		}

		private void verifyNestedReification(TriplesBlockNode root)
		{
			TriplesBlockNode[] triples = TripleUtils.FetchSubsequentTriples(root);
			List<TriplesBlockNode> triplesList = new List<TriplesBlockNode>(triples);

			Assert.AreEqual(11, triplesList.Count);

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new PrefixedNameNode() { Pn_Local = "x" },
				new VarNode("p"),
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new VarNode("v")));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new PrefixedNameNode() { Pn_Local = "a" }));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new BlankNode()));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new PrefixedNameNode() { Pn_Local = "b" }));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" }));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
				new PrefixedNameNode() { Pn_Local = "w" }));

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new BlankNode(),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" }));

			Assert.AreEqual(0, triplesList.Count);

			triplesList = new List<TriplesBlockNode>(triples);

			assertBlankNodesInLists(triplesList);
		}

		private void verifyEmptyReification(TriplesBlockNode root)
		{
			TriplesBlockNode[] triples = TripleUtils.FetchSubsequentTriples(root);
			List<TriplesBlockNode> triplesList = new List<TriplesBlockNode>(triples);

			Assert.AreEqual(1, triplesList.Count);

			Assert.IsTrue(TestUtils.TryRemoveFirstTriple(triplesList,
				new PrefixedNameNode() { Pn_Local = "x" },
				new VarNode("p"),
				new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" }));

			Assert.AreEqual(0, triplesList.Count);
		}

		private void assertBlankNodesInLists(List<TriplesBlockNode> triples)
		{
			INode subject, predicate, @object;

			for (int i = triples.Count - 1; i >= 0; i--)
			{
				subject = TripleUtils.GetSubject(triples[i]);
				predicate = TripleUtils.GetPredicate(triples[i]);
				@object = TripleUtils.GetObject(triples[i]);

				if (predicate.GetType() != typeof(IriRefNode))
				{
					followAndAssertList(@object as BlankNode, triples);
				}
			}
		}

		private void followAndAssertList(BlankNode blankNode, List<TriplesBlockNode> triples)
		{
			string first = "http://www.w3.org/1999/02/22-rdf-syntax-ns#first";
			string rest = "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest";
			string nil = "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil";

			TriplesBlockNode firstTriple = null, restTriple = null;
			BlankNode subject;
			IriRefNode predicate;
			INode @object;

			foreach (TriplesBlockNode triple in triples)
			{
				subject = TripleUtils.GetSubject(triple) as BlankNode;
				predicate = TripleUtils.GetPredicate(triple) as IriRefNode;

				if (subject != null && predicate != null)
				{
					if (subject.Value == blankNode.Value)
					{
						if (predicate.Value == first)
						{
							firstTriple = triple;
						}
						else if (predicate.Value == rest)
						{
							restTriple = triple;
						}
					}
				}
			}

			Assert.IsNotNull(firstTriple);
			Assert.IsNotNull(restTriple);

			@object = TripleUtils.GetObject(firstTriple);
			if (@object.GetType() == typeof(BlankNode) && (BlankNode)@object != null)
			{
				followAndAssertList((BlankNode)@object, triples);
			}

			@object = TripleUtils.GetObject(restTriple);
			if (@object.GetType() == typeof(BlankNode) && (BlankNode)@object != null)
			{
				followAndAssertList((BlankNode)@object, triples);
			}
			else
			{
				Assert.AreEqual(@object.GetType(), typeof(IriRefNode));
				Assert.AreEqual(((IriRefNode)@object).Value, nil);
			}
		}
	}
}
