﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Algebra.AstUtils.Visitor;
using SharQL.Algebra.Tests.Utils;

namespace SharQL.Algebra.AstUtils.Tests
{
	/// <summary>
	/// Summary description for PrefixedNamesExpansionTest
	/// </summary>
	[TestClass]
	public class PrefixedNamesExpansionTest
	{
		public PrefixedNamesExpansionTest()
		{
		}

		private TestContext testContextInstance;

		/// <summary>
		///Gets or sets the test context which provides
		///information about and functionality for the current test run.
		///</summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		//
		// You can use the following additional attributes as you write your tests:
		//
		// Use ClassInitialize to run code before running the first test in the class
		// [ClassInitialize()]
		// public static void MyClassInitialize(TestContext testContext) { }
		//
		// Use ClassCleanup to run code after all tests in a class have run
		// [ClassCleanup()]
		// public static void MyClassCleanup() { }
		//
		// Use TestInitialize to run code before running each test 
		// [TestInitialize()]
		// public void MyTestInitialize() { }
		//
		// Use TestCleanup to run code after each test has run
		// [TestCleanup()]
		// public void MyTestCleanup() { }
		//
		#endregion

		[TestMethod]
		public void TestExpansionOfPrefix()
		{
			QueryNode root = createQueryWithPrefixes();
			expandPrefixedNames(ref root);
			verifyExpansionOfPrefix(root);
		}

		[TestMethod]
		public void TestExpasionOfBase()
		{
			QueryNode root = createQueryWithBase();
			expandPrefixedNames(ref root);
			verifyExpansionOfBase(root);
		}

        [TestMethod]
        public void TestExpasionOfRelativeAndAbsoluteIRIs()
        {
            QueryNode root = createQueryWithRelativeAndAbsoluteIRIs();
            expandPrefixedNames(ref root);
            verifyExpansionOfRelativeAndAbsoluteIRIs(root);
        }

		[TestMethod]
		public void TestExpansionOfBaseAndPrefix()
		{
			QueryNode root = createQueryWithBaseAndPrefixes();
			expandPrefixedNames(ref root);
			verifyExpansionOfBaseAndPrefix(root);
		}

		[TestMethod]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void TestRelativeIriWhenBaseMissing()
		{
			QueryNode root = createQueryWithRelativeIriAndBaseMissing();
			expandPrefixedNames(ref root);
		}

		[TestMethod]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void TestPrefixedIriWhenPrefixMissing()
		{
			QueryNode root = createQueryWithPrefixedIriAndPrefixMissing();
			expandPrefixedNames(ref root);
		}

		/// <summary>
		/// <code>
		/// PREFIX ns: <http://example.org/ns#>
		/// PREFIX x:  <http://example.org/x/>
		/// 
		/// SELECT * WHERE { x:x ns:p ?v }
		/// </code>
		/// </summary>
		private QueryNode createQueryWithPrefixes()
		{
			return new QueryNode(
				new PrologueNode(
					null,
					new PrefixDeclListNode(
						new PrefixDeclNode(
							"ns",
							new IriRefNode() { Value = "http://example.org/ns#" }
						),
						new PrefixDeclListNode(
							new PrefixDeclNode(
								"x",
								new IriRefNode() { Value = "http://example.org/x/" }
							),
							null
						)
					)
				),
				new SelectQueryNode(
					new StarVarListNode() { IsStar = true },
					new WhereClauseNode(
						new GroupGraphPatternNode(
							new TriplesBlockNode(
								new TriplesSameSubjectNode(
									new PrefixedNameNode() { Pname_Ns = "x", Pn_Local = "x" },
									null,
									new PropertyListNode(
										new PrefixedNameNode() { Pname_Ns = "ns", Pn_Local = "p" },
										new ObjectListNode(
											new VarNode("v"),
											null
										),
										null
									)
								),
								null
							),
							null
						)
					)
				)
			);
		}

		/// <summary>
		/// <code>
		/// BASE <http://example.org/x/>
		/// 
		/// SELECT * WHERE { <x> <p> ?v } 
		/// </code>
		/// </summary>
		private QueryNode createQueryWithBase()
		{
			return new QueryNode(
				new PrologueNode(
					new PrefixDeclNode(
						new IriRefNode() { Value = "http://example.org/x/" }
					),
					null
				),
				new SelectQueryNode(
					new StarVarListNode() { IsStar = true },
					new WhereClauseNode(
						new GroupGraphPatternNode(
							new TriplesBlockNode(
								new TriplesSameSubjectNode(
									new IriRefNode() { Value = "x" },
									null,
									new PropertyListNode(
										new IriRefNode() { Value = "p" },
										new ObjectListNode(
											new VarNode("v"),
											null
										),
										null
									)
								),
								null
							),
							null
						)
					)
				)
			);
		}

        /// <summary>
        /// <code>
        /// BASE <http://example.org/x/>
        /// 
        /// SELECT * WHERE { <#skål> <http://www.skål.com/p> ?v } 
        /// </code>
        /// </summary>
        private QueryNode createQueryWithRelativeAndAbsoluteIRIs()
        {
            return new QueryNode(
                new PrologueNode(
                    new PrefixDeclNode(
                        new IriRefNode() { Value = "http://example.org/x/" }
                    ),
                    null
                ),
                new SelectQueryNode(
                    new StarVarListNode() { IsStar = true },
                    new WhereClauseNode(
                        new GroupGraphPatternNode(
                            new TriplesBlockNode(
                                new TriplesSameSubjectNode(
                                    new IriRefNode() { Value = "#skål" },
                                    null,
                                    new PropertyListNode(
                                        new IriRefNode() { Value = "http://www.skål.com/p" },
                                        new ObjectListNode(
                                            new VarNode("v"),
                                            null
                                        ),
                                        null
                                    )
                                ),
                                null
                            ),
                            null
                        )
                    )
                )
            );
        }

		/// <summary>
		/// <code>
		/// BASE <http://example.org/x/> 
		/// PREFIX : <#>
		/// 
		/// SELECT * WHERE { :x ?p ?v } 
		/// </code>
		/// </summary>
		private QueryNode createQueryWithBaseAndPrefixes()
		{
			return new QueryNode(
				new PrologueNode(
					new PrefixDeclNode(
						new IriRefNode() { Value = "http://example.org/x/" }
					),
					new PrefixDeclListNode(
						new PrefixDeclNode(
							"",
							new IriRefNode() { Value = "#" }
						),
						null
					)
				),
				new SelectQueryNode(
					new StarVarListNode() { IsStar = true },
					new WhereClauseNode(
						new GroupGraphPatternNode(
							new TriplesBlockNode(
								new TriplesSameSubjectNode(
									new PrefixedNameNode() { Pn_Local = "x" },
									null,
									new PropertyListNode(
										new VarNode("p"),
										new ObjectListNode(
											new VarNode("v"),
											null
										),
										null
									)
								),
								null
							),
							null
						)
					)
				)
			);
		}

		/// <summary>
		/// <code>
		/// PREFIX foo: <http://example.org/>
		/// 
		/// SELECT *
		/// {
		///		?a ?b bar:error .
		/// }
		/// </code>
		/// </summary>
		private QueryNode createQueryWithPrefixedIriAndPrefixMissing()
		{
			return new QueryNode(
				new PrologueNode(
					null,
					new PrefixDeclListNode(
						new PrefixDeclNode(
							"foo",
							new IriRefNode() { Value = "http://example.org/" }
						),
						null
					)
				),
				new SelectQueryNode(
					new StarVarListNode() { IsStar = true },
					new WhereClauseNode(
						new GroupGraphPatternNode(
							new TriplesBlockNode(
								new TriplesSameSubjectNode(
									new VarNode("a"),
									null,
									new PropertyListNode(
										new VarNode("b"),
										new ObjectListNode(
											new PrefixedNameNode() { Pname_Ns = "bar", Pn_Local = "error" },
											null
										),
										null
									)
								),
								null
							),
							null
						)
					)
				)
			);
		}

		/// <summary>
		/// <code>
		/// PREFIX foo: <http://example.org/>
		/// 
		/// SELECT *
		/// {
		///		?a ?b <a/b/c> .
		/// }
		/// </code>
		/// </summary>
		private QueryNode createQueryWithRelativeIriAndBaseMissing()
		{
			return new QueryNode(
				new PrologueNode(
					null,
					new PrefixDeclListNode(
						new PrefixDeclNode(
							"foo",
							new IriRefNode() { Value = "http://example.org/" }
						),
						null
					)
				),
				new SelectQueryNode(
					new StarVarListNode() { IsStar = true },
					new WhereClauseNode(
						new GroupGraphPatternNode(
							new TriplesBlockNode(
								new TriplesSameSubjectNode(
									new VarNode("a"),
									null,
									new PropertyListNode(
										new VarNode("b"),
										new ObjectListNode(
											new IriRefNode(){Value = "a/b/c"},
											null
										),
										null
									)
								),
								null
							),
							null
						)
					)
				)
			);
		}

		private void expandPrefixedNames(ref QueryNode root)
		{
            ExpansionVisitor visitor = new ExpansionVisitor();
            root.Accept(visitor);
		}

		private void verifyExpansionOfPrefix(QueryNode root)
		{
			TriplesBlockNode triple = root.Children[1].Children[1].Children[0].Children[0] as TriplesBlockNode;
			
			assertTripleContent(triple,
				new IriRefNode() { Value = "http://example.org/x/x" },
				new IriRefNode() { Value = "http://example.org/ns#p" },
				new VarNode("v"));
		}

		private void verifyExpansionOfBase(QueryNode root)
		{
			TriplesBlockNode triple = root.Children[1].Children[1].Children[0].Children[0] as TriplesBlockNode;
			assertTripleContent(triple,
				new IriRefNode() { Value = "http://example.org/x/x" },
				new IriRefNode() { Value = "http://example.org/x/p" },
				new VarNode("v"));
		}

        private void verifyExpansionOfRelativeAndAbsoluteIRIs(QueryNode root)
        {
            TriplesBlockNode triple = root.Children[1].Children[1].Children[0].Children[0] as TriplesBlockNode;
            assertTripleContent(triple,
                new IriRefNode() { Value = "http://example.org/x/#skål" },
                new IriRefNode() { Value = "http://www.skål.com/p" },
                new VarNode("v"));
        }

		private void verifyExpansionOfBaseAndPrefix(QueryNode root)
		{
			TriplesBlockNode triple = root.Children[1].Children[1].Children[0].Children[0] as TriplesBlockNode;
			assertTripleContent(triple,
				new IriRefNode() { Value = "http://example.org/x/#x" },
				new VarNode("p"),
				new VarNode("v"));
		}

		private void assertTripleContent(TriplesBlockNode triple, INode subject, INode predicate, INode @object)
		{
			INode realSubject = TripleUtils.GetSubject(triple), realPredicate = TripleUtils.GetPredicate(triple), realObject = TripleUtils.GetObject(triple);

			Assert.AreEqual(subject.GetType(), realSubject.GetType());
			Assert.AreEqual(predicate.GetType(), realPredicate.GetType());
			Assert.AreEqual(@object.GetType(), realObject.GetType());

			assertSameValues(subject, realSubject);
			assertSameValues(predicate, realPredicate);
			assertSameValues(@object, realObject);
		}

		private void assertSameValues(INode node, INode realNode)
		{
			if (node.GetType() == typeof(PrefixedNameNode) && realNode.GetType() == typeof(PrefixedNameNode))
			{
				PrefixedNameNode pnnNode = (PrefixedNameNode)node;
				PrefixedNameNode pnnRealNode = (PrefixedNameNode)realNode;

				Assert.AreEqual(pnnNode.Pn_Local, pnnRealNode.Pn_Local);
				Assert.AreEqual(pnnNode.Pname_Ns, pnnRealNode.Pname_Ns);
			}
			else if (node.GetType() == typeof(VarNode) && realNode.GetType() == typeof(VarNode))
			{
				VarNode vnNode = (VarNode)node;
				VarNode vnRealNode = (VarNode)realNode;

				Assert.AreEqual(vnNode.Value, vnRealNode.Value);
			}
			else if (node.GetType() == typeof(IriRefNode) && realNode.GetType() == typeof(IriRefNode))
			{
				IriRefNode irnNode = (IriRefNode)node;
				IriRefNode irnRealNode = (IriRefNode)realNode;

				Assert.AreEqual(irnNode.Value, irnRealNode.Value);
			}
		}
	}
}
