﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Algebra.AstUtils.Visitor;

namespace SharQL.Algebra.QueryStructure.Tests
{
	/// <summary>
	/// Summary description for WhereClauseTransformTest
	/// </summary>
	[TestClass]
	public class WhereClauseTransformTest
	{
		public WhereClauseTransformTest()
		{
		}

		private TestContext testContextInstance;

		/// <summary>
		///Gets or sets the test context which provides
		///information about and functionality for the current test run.
		///</summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		//
		// You can use the following additional attributes as you write your tests:
		//
		// Use ClassInitialize to run code before running the first test in the class
		// [ClassInitialize()]
		// public static void MyClassInitialize(TestContext testContext) { }
		//
		// Use ClassCleanup to run code after all tests in a class have run
		// [ClassCleanup()]
		// public static void MyClassCleanup() { }
		//
		// Use TestInitialize to run code before running each test 
		// [TestInitialize()]
		// public void MyTestInitialize() { }
		//
		// Use TestCleanup to run code after each test has run
		// [TestCleanup()]
		// public void MyTestCleanup() { }
		//
		#endregion

		[TestMethod]
		public void TestWhereClauseOneTriple()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s ?p ?o }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(BasicGraphPattern));

			BasicGraphPattern bgp = (BasicGraphPattern)result;
			Assert.AreEqual(1, bgp.Triples.Count);

			Triple triple = bgp.Triples[0];
			Assert.IsInstanceOfType(triple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(triple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(triple.Object, typeof(Resource));
			Assert.AreEqual(triple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(triple.Predicate.Type, ResourceType.Variable);
			Assert.AreEqual(triple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(triple.Subject.Value, "s");
			Assert.AreEqual(triple.Predicate.Value, "p");
			Assert.AreEqual(triple.Object.Value, "o");
		}

		[TestMethod]
		public void TestWhereClauseTwoTriples()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s <a:b> ?v1 ; <a:c> ?v2 }");
			INode ast = p.AstRoot;
			ast.Accept(new TriplesSameSubjectResolverVisitor());
			WhereClauseNode whereClause = getWhereClauseFromQuery(ast);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(Join));

			Join join = (Join)result;

			Assert.IsInstanceOfType(join.FirstPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(join.SecondPattern, typeof(BasicGraphPattern));

			BasicGraphPattern bgp1 = (BasicGraphPattern)join.FirstPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)join.SecondPattern;

			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Assert.IsInstanceOfType(firstTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Object, typeof(Resource));
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
		}

		[TestMethod]
		public void TestWhereClauseOneUnion()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { { ?s <a:b> ?v1 } UNION {?s <a:c> ?v2 } }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(Union));

			Union union = (Union)result;
			Assert.IsInstanceOfType(union.FirstPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(union.SecondPattern, typeof(BasicGraphPattern));

			BasicGraphPattern bgp1 = (BasicGraphPattern)union.FirstPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)union.SecondPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Assert.IsInstanceOfType(firstTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Object, typeof(Resource));
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
		}

		[TestMethod]
		public void TestWhereClauseTwoUnions()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { { ?s <a:b> ?v1 } UNION {?s <a:c> ?v2 } UNION {?s <a:d> ?v3 } }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(Union));

			Union outerUnion = (Union)result;
			Assert.IsInstanceOfType(outerUnion.FirstPattern, typeof(Union));
			Assert.IsInstanceOfType(outerUnion.SecondPattern, typeof(BasicGraphPattern));

			Union innerUnion = (Union)outerUnion.FirstPattern;
			Assert.IsInstanceOfType(innerUnion.FirstPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(innerUnion.SecondPattern, typeof(BasicGraphPattern));

			BasicGraphPattern bgp1 = (BasicGraphPattern)innerUnion.FirstPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)innerUnion.SecondPattern;
			BasicGraphPattern bgp3 = (BasicGraphPattern)outerUnion.SecondPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);
			Assert.AreEqual(1, bgp3.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Triple thirdTriple = bgp3.Triples[0];

			Assert.IsInstanceOfType(firstTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Object, typeof(Resource));
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
			Assert.AreEqual(thirdTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(thirdTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Subject.Value, "s");
			Assert.AreEqual(thirdTriple.Predicate.Value, "a:d");
			Assert.AreEqual(thirdTriple.Object.Value, "v3");
		}

		[TestMethod]
		public void TestWhereClauseOneOptional()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s <a:b> ?v1 OPTIONAL {?s <a:c> ?v2 } }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(LeftJoin));

			LeftJoin optional = (LeftJoin)result;
			Assert.IsInstanceOfType(optional.LeftPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.ConstraintExpression, typeof(BooleanLiteralNode));

			BooleanLiteralNode constraint = (BooleanLiteralNode)optional.ConstraintExpression;
			Assert.IsTrue(constraint.Value);

			BasicGraphPattern bgp1 = (BasicGraphPattern)optional.LeftPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)optional.RightPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);
		
			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
		}

		[TestMethod]
		public void TestWhereClauseTwoOptionals()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s <a:b> ?v1 OPTIONAL {?s <a:c> ?v2 } OPTIONAL { ?s <a:d> ?v3 } }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(LeftJoin));
			
			LeftJoin outerOptional = (LeftJoin)result;
			Assert.IsInstanceOfType(outerOptional.LeftPattern, typeof(LeftJoin));
			Assert.IsInstanceOfType(outerOptional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(outerOptional.ConstraintExpression, typeof(BooleanLiteralNode));

			LeftJoin innerOptional = (LeftJoin)outerOptional.LeftPattern;
			Assert.IsInstanceOfType(innerOptional.LeftPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(innerOptional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(innerOptional.ConstraintExpression, typeof(BooleanLiteralNode));

			BooleanLiteralNode constraint1 = (BooleanLiteralNode)outerOptional.ConstraintExpression;
			BooleanLiteralNode constraint2 = (BooleanLiteralNode)innerOptional.ConstraintExpression;
			Assert.IsTrue(constraint1.Value);
			Assert.IsTrue(constraint2.Value);

			BasicGraphPattern bgp1 = (BasicGraphPattern)innerOptional.LeftPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)innerOptional.RightPattern;
			BasicGraphPattern bgp3 = (BasicGraphPattern)outerOptional.RightPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);
			Assert.AreEqual(1, bgp3.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Triple thirdTriple = bgp3.Triples[0];

			Assert.IsInstanceOfType(firstTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Object, typeof(Resource));
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
			Assert.AreEqual(thirdTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(thirdTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Subject.Value, "s");
			Assert.AreEqual(thirdTriple.Predicate.Value, "a:d");
			Assert.AreEqual(thirdTriple.Object.Value, "v3");
		}
		
		[TestMethod]
		public void TestWhereClauseOptionalFilter()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s <a:b> ?v1 OPTIONAL {?s <a:c> ?v2 FILTER(?v1<3) } } ");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(LeftJoin));

			LeftJoin optional = (LeftJoin)result;
			Assert.IsInstanceOfType(optional.LeftPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.ConstraintExpression, typeof(ConditionalOrExpressionNode));
	
			BasicGraphPattern bgp1 = (BasicGraphPattern)optional.LeftPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)optional.RightPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
		}

		[TestMethod]
		public void TestWhereClauseUnionOptional()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { {?s <a:b> ?v1} UNION {?s <a:c> ?v2} OPTIONAL {?s <a:d> ?v3} } ");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(LeftJoin));

			LeftJoin optional = (LeftJoin)result;
			Assert.IsInstanceOfType(optional.LeftPattern, typeof(Union));
			Assert.IsInstanceOfType(optional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.ConstraintExpression, typeof(BooleanLiteralNode));

			BooleanLiteralNode constraint = (BooleanLiteralNode)optional.ConstraintExpression;
			Assert.IsTrue(constraint.Value);

			Union union = (Union)optional.LeftPattern;
			Assert.IsInstanceOfType(union.FirstPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(union.SecondPattern, typeof(BasicGraphPattern));

			BasicGraphPattern bgp1 = (BasicGraphPattern)union.FirstPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)union.SecondPattern;
			BasicGraphPattern bgp3 = (BasicGraphPattern)optional.RightPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);
			Assert.AreEqual(1, bgp3.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Triple thirdTriple = bgp3.Triples[0];

			Assert.IsInstanceOfType(firstTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(firstTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(secondTriple.Object, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Subject, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Predicate, typeof(Resource));
			Assert.IsInstanceOfType(thirdTriple.Object, typeof(Resource));
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
			Assert.AreEqual(thirdTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(thirdTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(thirdTriple.Subject.Value, "s");
			Assert.AreEqual(thirdTriple.Predicate.Value, "a:d");
			Assert.AreEqual(thirdTriple.Object.Value, "v3");
		}

		[TestMethod]
		public void TestWhereClauseFilterOptional()
		{
			Parser p = new Parser();
			p.Parse("SELECT * { ?s <a:b> ?v1 FILTER (?v1 < 3 ) OPTIONAL {?s <a:c> ?v2} }");
			WhereClauseNode whereClause = getWhereClauseFromQuery(p.AstRoot);
			Assert.IsNotNull(whereClause);

			Pattern result = WhereClauseTransformer.Transform(whereClause);
			Assert.IsNotNull(result);
			Assert.IsInstanceOfType(result, typeof(Filter));

			Filter filter = (Filter)result;
			Assert.IsInstanceOfType(filter.ConstraintExpression, typeof(ConditionalOrExpressionNode));
			Assert.IsInstanceOfType(filter.Pattern, typeof(LeftJoin));

			LeftJoin optional = (LeftJoin)filter.Pattern;
			Assert.IsInstanceOfType(optional.LeftPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.RightPattern, typeof(BasicGraphPattern));
			Assert.IsInstanceOfType(optional.ConstraintExpression, typeof(BooleanLiteralNode));

			BooleanLiteralNode constraint = (BooleanLiteralNode)optional.ConstraintExpression;
			Assert.IsTrue(constraint.Value);

			BasicGraphPattern bgp1 = (BasicGraphPattern)optional.LeftPattern;
			BasicGraphPattern bgp2 = (BasicGraphPattern)optional.RightPattern;
			Assert.AreEqual(1, bgp1.Triples.Count);
			Assert.AreEqual(1, bgp2.Triples.Count);

			Triple firstTriple = bgp1.Triples[0];
			Triple secondTriple = bgp2.Triples[0];
			Assert.AreEqual(firstTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(firstTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(firstTriple.Subject.Value, "s");
			Assert.AreEqual(firstTriple.Predicate.Value, "a:b");
			Assert.AreEqual(firstTriple.Object.Value, "v1");
			Assert.AreEqual(secondTriple.Subject.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Predicate.Type, ResourceType.Iri);
			Assert.AreEqual(secondTriple.Object.Type, ResourceType.Variable);
			Assert.AreEqual(secondTriple.Subject.Value, "s");
			Assert.AreEqual(secondTriple.Predicate.Value, "a:c");
			Assert.AreEqual(secondTriple.Object.Value, "v2");
		}

		private WhereClauseNode getWhereClauseFromQuery(INode query)
		{
			return ((SelectQueryNode)query.Children[1]).WhereClause as WhereClauseNode;
		}
	}
}
