﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharQL.Ast.Visitor;
using SharQL.Ast;
using SharQL.Ast.Nodes;

namespace SharQL.Algebra.AstUtils.Visitor
{
	/// <summary>
	/// Validates the blank node labels, throwing an exception if a
	/// blank node labels have been used across basic graph patterns.
	/// </summary>
	/// <remarks>
	/// Must be applied to children before parent.
	/// </remarks>
	internal class BlankNodeLabelValidatorVisitor : IVisitor
	{
		private Dictionary<int, List<string>> nodeLabelMap = new Dictionary<int, List<string>>();
		private int scope = 0;

		#region IVisitor Members

		/// <summary>Indexer required for visitor support.</summary>
		/// <param name="nodeType">Ignored</param>
		/// <returns>A delegate for validating the use of blank node
		/// labels.</returns>
		public Action<INode> this[Type nodeType]
		{
			get
			{
				if (nodeType == typeof(QueryNode))
				{
					return validateBlankNodeLabels;
				}
				else if (nodeType == typeof(BlankNode))
				{
					return registerBlankNode;
				}
				else if (nodeType == typeof(OptionalGraphPatternNode)
					|| nodeType == typeof(GraphGraphPatternNode)
					|| nodeType == typeof(GroupGraphPatternNode))
				{
					return incrementScope;
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		private void validateBlankNodeLabels(INode node)
		{
			List<string>[] bnodeLabelLists = nodeLabelMap.Values.ToArray();

			for (int i = 0; i < bnodeLabelLists.Length - 1; i++)
			{
				for (int j = i + 1; j < bnodeLabelLists.Length; j++)
				{
					if (bnodeLabelLists[i].Intersect(bnodeLabelLists[j]).Count() > 0)
					{
						throw new InvalidOperationException("A label can be used in only a single basic graph pattern in any query.");
					}
				}
			}
		}

		private void registerBlankNode(INode node)
		{
			BlankNode bnode = (BlankNode)node;
			addNodeLabel(bnode.Value);
		}

		private void addNodeLabel(string label)
		{
			if (!nodeLabelMap.ContainsKey(scope))
			{
				nodeLabelMap[scope] = new List<string>();
			}

			nodeLabelMap[scope].Add(label);
		}

		private void incrementScope(INode node)
		{
			scope++;
		}
	}
}
