﻿using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Ast.Visitor;

namespace SharQL.Algebra.AstUtils.Visitor
{
    internal class ExpansionVisitor : IVisitor
    {
        private Dictionary<string, string> prefixes;
        private string basePrefix = null;

        /// <summary>Indexer required for visitor support.</summary>
        /// <param name="nodeType">Ignored</param>
        /// <returns>A delegate for expanding bases and prefixes.</returns>
        public Action<INode> this[Type nodeType]
        {
            get
            {
                if (nodeType == typeof(IriRefNode))
                {
                    return expandRelativeIRIs;
                }
                else if (nodeType == typeof(PrefixedNameNode))
                {
                    return expandPrefixedName;
                }

                return null;
            }
        }

        private void expandRelativeIRIs(INode node)
        {
            // Only expand relative IRIs, leave absolute ones alone
            if (!Regex.IsMatch((node as IriRefNode).Value, "^[^:]+:"))
            //if (!Uri.IsWellFormedUriString((node as IriRefNode).Value, UriKind.Absolute))
            {
                (node as IriRefNode).Value = getBasePrefix(node) + (node as IriRefNode).Value;
            }
        }

        private void expandPrefixedName(INode node)
        {
            if (prefixes == null)
            {
                mapPrefixes(node);
            }

            // If node namespace value is null, go with "" to match the default/empty prefix
            string prefix = (node as PrefixedNameNode).Pname_Ns ?? "";

            if (!prefixes.ContainsKey(prefix))
            {
                throw new ArgumentOutOfRangeException("PREFIX", string.Format("Undefined prefix '{0}'.", (node as PrefixedNameNode).Pname_Ns));
            }

            // Create replacement IriRefNode concatenating resolved prefix and local name
            INode iriRef = new IriRefNode() { Value = prefixes[prefix] + (node as PrefixedNameNode).Pn_Local };

            Util.ReplaceNode(node, iriRef);

            // Expand IriRefNode in case of relative IRI
            expandRelativeIRIs(iriRef);
        }

        private string getBasePrefix(INode node)
        {
            if (prefixes == null)
            {
                mapPrefixes(node);
            }

            if (basePrefix == null)
            {
                throw new ArgumentOutOfRangeException("BASE", string.Format("No base prefix declaration found while expanding relative IRI '{0}'.", (node as IriRefNode).Value));
            }

            return basePrefix;
        }

        private void mapPrefixes(INode node)
        {
            prefixes = new Dictionary<string, string>();

            // Locate root node
            while (node.Parent != null)
            {
                node = node.Parent;
            }

            // Check for prologue node
            INode prologueNode = (node as QueryNode).Children[0];
            if (prologueNode != null)
            {
                // Fetch base prefix, if present
                if (prologueNode.Children[0] != null)
                {
                    basePrefix = (prologueNode.Children[0].Children[0] as IriRefNode).Value;
                }

                // Enumerate prefix declarations
                INode currentNode = prologueNode.Children[1];
                while (currentNode != null)
                {
                    PrefixDeclNode prefixDecl = currentNode.Children[0] as PrefixDeclNode;

                    // Map prefix declaration
                    prefixes[prefixDecl.Namespace] = prefixDecl.IriRef.Value;

                    currentNode = currentNode.Children[1];
                }
            }
        }
    }
}
