﻿using System;
using System.Collections.Generic;
using System.Linq;

using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Ast.Visitor;

namespace SharQL.Algebra.AstUtils.Visitor
{
    internal class ListReificationVisitor : IVisitor
    {
        /// <summary>Indexer required for visitor support.</summary>
        /// <param name="nodeType">Ignored</param>
        /// <returns>A delegate for translating syntactic sugar for triple block
        /// lists into its core representation.</returns>
        public Action<INode> this[Type nodeType]
        {
            get
            {
                if (nodeType == typeof(CollectionNode))
                {
                    return reificateList;
                }
				else if (nodeType == typeof(ObjectListNode))
				{
					return reificateEmptyList;
				}

                return null;
            }
        }

		private void reificateEmptyList(INode node)
		{
			ObjectListNode oln = (ObjectListNode)node;
			if (oln.Children.All(n => n == null))
			{
				IriRefNode iri = new IriRefNode() { Value = @"http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" };
				oln.Children[0] = iri;
				iri.Parent = oln;
			}
		}

        private void reificateList(INode node)
        {
            // Collect list elements
            List<INode> elements = new List<INode>();
            INode tripleBlock = node.Children[0];
            while (tripleBlock != null && tripleBlock.Children.Count > 0)
            {
                elements.Add(tripleBlock.Children[0]);
                tripleBlock = tripleBlock.Children.Count == 2 ? tripleBlock.Children[1] : null;
            }

            // Find anchestor TriplesBlockNode
            tripleBlock = node;
            while (tripleBlock.GetType() != typeof(TriplesBlockNode))
            {
                tripleBlock = tripleBlock.Parent;
            }

            INode initialBlankNode = new BlankNode() { Value = Guid.NewGuid().ToString() };
            INode currentBlankNode = initialBlankNode;
            INode nextBlankNode = new BlankNode() { Value = Guid.NewGuid().ToString() };

            INode root = tripleBlock;

            // Build collection subtree
            foreach (INode element in elements)
            {
                INode firstTriple = new TriplesBlockNode(
                    new TriplesSameSubjectNode(
                        currentBlankNode as NodeBase, 
                        null,
                        new PropertyListNode(
                            new IriRefNode() { Value = "http://www.w3.org/1999/02/22-rdf-syntax-ns#first" },
                            new ObjectListNode(
                                element as NodeBase, 
                                null
                            ),
                            null
                        )
                    ),
                    null
                );

                Util.InsertTripleBlock(root, firstTriple);

                INode restTriple = new TriplesBlockNode(
                    new TriplesSameSubjectNode(
                        currentBlankNode as NodeBase,
                        null,
                        new PropertyListNode(
                            new IriRefNode() { Value = "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest" },
                            new ObjectListNode(
                                (element != elements.Last() ? 
                                    nextBlankNode as NodeBase :
                                    new IriRefNode() { Value = "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil" }),
                                null
                            ),
                            null
                        )
                    ),
                    null
                );

                Util.InsertTripleBlock(firstTriple, restTriple);
                root = restTriple;

                // Increment blank nodes
                currentBlankNode = nextBlankNode;
                nextBlankNode = new BlankNode() { Value = Guid.NewGuid().ToString() };
            }

            // Replace CollectionNode with new AST subtree
            if (node.Parent is TriplesSameSubjectNode &&
                    node.Parent.Children[2] == null)
            {
                if (tripleBlock.Parent is GroupGraphPatternNode)
                {
                    tripleBlock.Parent.Children[0] = tripleBlock.Children[1];
                    tripleBlock.Children[1].Parent = tripleBlock.Parent;
                }
                else if (tripleBlock.Parent is TriplesBlockNode)
                {
                    tripleBlock.Parent.Children[1] = tripleBlock.Children[1];
                    tripleBlock.Children[1].Parent = tripleBlock.Parent;
                }
            }
            else
            {
                Util.ReplaceNode(node, initialBlankNode.Clone() as INode);
            }
        }
    }
}
