﻿using System;
using System.Globalization;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Ast.Visitor;

namespace SharQL.Algebra.AstUtils.Visitor
{
	internal class LiteralExplicatorVisitor : IVisitor
	{
		/// <summary>
		/// Gets a delegate for handling the specified node type. Used by the
		/// visitor pattern.
		/// </summary>
		/// <value>A delegate for handling the specified node type.</value>
		/// <returns>A delegate for translating custom literal AST node into
		/// general RDF literal AST node.</returns>
		public Action<INode> this[Type nodeType]
		{
			get
			{
				if (nodeType == typeof(NumericLiteralNode) ||
					nodeType == typeof(BooleanLiteralNode))
				{
					return explicateNode;
				}
				else
				{
					return null;
				}
			}
		}

		/// <summary>
		/// Explicates the node. If the node is a <see cref="NumericLiteralNode"/>
		/// or a <see cref="BooleanLiteralNode"/>, it is replaced with an
		/// <see cref="RdfLiteralNode"/> containing the same data, but on explicit
		/// form.
		/// </summary>
		/// <param name="node">The node to explicate.</param>
		private void explicateNode(INode node)
		{
			INode newNode = null;

			if (node is NumericLiteralNode)
			{
				newNode = explicateNumericLiteralNode((NumericLiteralNode)node);
			}
			else if (node is BooleanLiteralNode)
			{
				newNode = explicateBooleanLiteralNode((BooleanLiteralNode)node);
			}

			if (newNode != null)
			{
				Util.ReplaceNode(node, newNode);
			}
		}

		/// <summary>
		/// Explicates the specified <see cref="NumericLiteralNode"/>.
		/// </summary>
		/// <param name="node">The node to explicate.</param>
		/// <returns>The explicit form of the node, as an <see cref="RdfLiteralNode"/>.</returns>
		private RdfLiteralNode explicateNumericLiteralNode(NumericLiteralNode node)
		{
			switch (node.Type)
			{
				case NumericLiteralNode.LiteralType.Int:
					return new RdfLiteralNode(
						new StringLiteralNode { Value = node.IntValue.ToString() },
						new IriRefNode { Value = "http://www.w3.org/2001/XMLSchema#integer" }
					);

				case NumericLiteralNode.LiteralType.Double:
					return new RdfLiteralNode(
						new StringLiteralNode { Value = node.DoubleValue.ToString(CultureInfo.InvariantCulture) },
						new IriRefNode { Value = "http://www.w3.org/2001/XMLSchema#double" }
					);

				case NumericLiteralNode.LiteralType.Decimal:
					return new RdfLiteralNode(
						new StringLiteralNode { Value = node.DecimalValue.ToString(CultureInfo.InvariantCulture) },
						new IriRefNode { Value = "http://www.w3.org/2001/XMLSchema#decimal" }
					);

				default: return null;
			}
		}

		/// <summary>
		/// Explicates the specified <see cref="BooleanLiteralNode"/>.
		/// </summary>
		/// <param name="node">The node to explicate.</param>
		/// <returns>The explicit form of the node, as an <see cref="RdfLiteralNode"/>.</returns>
		private RdfLiteralNode explicateBooleanLiteralNode(BooleanLiteralNode node)
		{
			return new RdfLiteralNode(
				new StringLiteralNode { Value = node.Value ? "true" : "false" },
				new IriRefNode { Value = "http://www.w3.org/2001/XMLSchema#boolean" }
			);
		}
	}
}
