﻿using System;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Algebra.QueryStructure;
using SharQL.Algebra.AstUtils.Visitor;
using SharQL.Ast.Visitor;

namespace SharQL.Algebra
{
	/// <summary>
	/// Contains functionality for transforming syntax trees into <see cref="Query"/> objects.
	/// </summary>
	public static class QueryTransformer
	{
		/// <summary>
		/// GUID-based prefix for distinguishing blank nodes from variables.
		/// </summary>
		public static readonly string BlankNodeVariablePrefix = "k0k0_"; //"A1527DF1_8030_4107_8B68_FF3AA59CD765_";

		/// <summary>
		/// Translates the specified query syntax tree.
		/// </summary>
		/// <param name="queryNode">The query node root of the syntax tree.</param>
		/// <returns>A <see cref="Query"/> object representing the query.</returns>
		public static Query Transform(QueryNode queryNode)
		{
			queryNode.PrepareForTransformation();
			Query query = CreateQueryFromPrologueNode(queryNode.Children[0] as PrologueNode);

			if (queryNode.Children[1] is SelectQueryNode)
			{
				query.Description = SelectQueryTransformer.TransformSelectQueryNode(queryNode.Children[1] as SelectQueryNode);
			}
			else if (queryNode.Children[1] is AskQueryNode)
			{
				throw new ArgumentException("ASK queries are currently not supported.", "query");
			}
			else if (queryNode.Children[1] is DescribeQueryNode)
			{
				throw new ArgumentException("DESCRIBE queries are currently not supported.", "query");
			}
			else if (queryNode.Children[1] is ConstructQueryNode)
			{
				throw new ArgumentException("CONSTRUCT queries are currently not supported.", "query");
			}
			else
			{
				throw new ArgumentException("Unrecognized syntax tree.", "query");
			}

			return query;
		}

		/// <summary>
		/// Prepares the AST for transformation.
		/// </summary>
		/// <param name="queryNode">The root node of the AST.</param>
		private static void PrepareForTransformation(this QueryNode queryNode)
		{
			queryNode.Accept(new BlankNodeLabelValidatorVisitor(), VisitorOrder.ChildrenBeforeParent);
			queryNode.Accept(new ExpansionVisitor(), VisitorOrder.ChildrenBeforeParent);
			queryNode.Accept(new LiteralExplicatorVisitor(), VisitorOrder.ChildrenBeforeParent);

			queryNode.Accept(new ListReificationVisitor(), VisitorOrder.ChildrenBeforeParent);
			queryNode.Accept(new BlankNodeReificationVisitor(), VisitorOrder.ChildrenBeforeParent);
			queryNode.Accept(new TriplesSameSubjectResolverVisitor(), VisitorOrder.ParentBeforeChildren);
		}

		/// <summary>
		/// Creates a <see cref="Query"/> object based on the specified prologue node.
		/// </summary>
		/// <param name="prologueNode">The prologue node on which the result is based.</param>
		/// <returns>A <see cref="Query"/> object based on the specified prologue node.</returns>
		private static Query CreateQueryFromPrologueNode(PrologueNode prologueNode)
		{
			Query query = new Query();

			if (prologueNode != null)
			{
				if (prologueNode.Children[0] != null)
				{
					query.Base = (prologueNode.Children[0].Children[0] as IriRefNode).Value;
				}

				INode currentNode = prologueNode.Children[1];
				while (currentNode != null)
				{
					PrefixDeclNode prefixDecl = currentNode.Children[0] as PrefixDeclNode;
					query.Prefixes[prefixDecl.Namespace] = prefixDecl.IriRef.Value;
					currentNode = currentNode.Children[1];
				}
			}

			return query;
		}
	}
}