﻿using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Algebra.QueryStructure;

namespace SharQL.Algebra
{
	/// <summary>
	/// Contains functionality for transforming <see cref="SelectQueryNode"/> objects
	/// into <see cref="SelectQueryDescription"/> objects.
	/// </summary>
    internal static class SelectQueryTransformer
    {
		/// <summary>
		/// Transforms the specified <see cref="SelectQueryNode"/> into a <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="selectQueryNode">The node to transform.</param>
		/// <returns>The resulting <see cref="SelectQueryDescription"/>.</returns>
		public static SelectQueryDescription TransformSelectQueryNode(SelectQueryNode selectQueryNode)
        {
            SelectQueryDescription description = new SelectQueryDescription()
				.transformDistinctReducedModifierNode(selectQueryNode.DistinctReducedModifier as DistinctReducedModifierNode)
				.transformStarVarListNode(selectQueryNode.StarVarList as StarVarListNode)
				.transformDatasetClauseListNode(selectQueryNode.DatasetClauseList as DatasetClauseListNode)
				.transformWhereClauseNode(selectQueryNode.WhereClause as WhereClauseNode)
				.transformSolutionModifierNode(selectQueryNode.SolutionModifier as SolutionModifierNode);

			return description;
        }

		/// <summary>
		/// Transforms the specified <see cref="DistinctReducedModifierNode"/> and adds it to
		/// the <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="description">The description to which the transformed object is added.</param>
		/// <param name="distinctReducedModifierNode">The node to transform.</param>
		/// <returns>A reference to <paramref name="description"/> (for chaining calls).</returns>
		private static SelectQueryDescription transformDistinctReducedModifierNode(this SelectQueryDescription description, DistinctReducedModifierNode distinctReducedModifierNode)
        {
            if (distinctReducedModifierNode != null)
            {
                if (distinctReducedModifierNode.Modifier == DistinctReducedModifierNode.Modifiers.Distinct)
                {
                    description.DistinctReducedModifier = DistinctReducedModifiers.Distinct;
                }
                else
                {
                    description.DistinctReducedModifier = DistinctReducedModifiers.Reduced;
                }
            }
            else
            {
                description.DistinctReducedModifier = DistinctReducedModifiers.None;
            }

			return description;
        }

		/// <summary>
		/// Transforms the specified <see cref="StarVarListNode"/> and adds it to the
		/// <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="description">The description to which the transformed object is added.</param>
		/// <param name="starVarListNode">The node to transform.</param>
		/// <returns>A reference to <paramref name="description"/> (for chaining calls).</returns>
		private static SelectQueryDescription transformStarVarListNode(this SelectQueryDescription description, StarVarListNode starVarListNode)
        {
            if (!starVarListNode.IsStar)
            {
                description.Variables = Util.FlattenRecursiveList<string>(starVarListNode.Children[0], starVarListNode.Children[1], n => (n as VarNode).Value);
            }

			return description;
        }

		/// <summary>
		/// Transforms the specified <see cref="DatasetClauseListNode"/> and adds it to the
		/// <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="description">The description to which the transformed object is added.</param>
		/// <param name="datasetClauseListNode">The node to transform.</param>
		/// <returns>A reference to <paramref name="description"/> (for chaining calls).</returns>
		private static SelectQueryDescription transformDatasetClauseListNode(this SelectQueryDescription description, DatasetClauseListNode datasetClauseListNode)
        {
            if (datasetClauseListNode != null)
            {
				Util.TraverseRecursiveList<SelectQueryDescription>(description, datasetClauseListNode.Children[0], datasetClauseListNode.Children[1], (n, d) =>
				{
					if (n.Children[0].Children[0] is DefaultGraphClauseNode)
					{
						d.DefaultGraphs.Add((n.Children[0].Children[0].Children[0].Children[0] as IriRefNode).Value);
					}
					else
					{
						d.NamedGraphs.Add((n.Children[0].Children[0].Children[0].Children[0] as IriRefNode).Value);
					}
				});
            }

			return description;
		}

		/// <summary>
		/// Transforms the specified <see cref="WhereClauseNode"/> and adds it to the
		/// <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="description">The description to which the transformed object is added.</param>
		/// <param name="whereClauseNode">The node to transform.</param>
		/// <returns>A reference to <paramref name="description"/> (for chaining calls).</returns>
		private static SelectQueryDescription transformWhereClauseNode(this SelectQueryDescription description, WhereClauseNode whereClauseNode)
        {
            description.WhereClause = WhereClauseTransformer.Transform(whereClauseNode);

			return description;
		}

		/// <summary>
		/// Transforms the specified <see cref="SolutionModifierNode"/> and adds it to the
		/// <see cref="SelectQueryDescription"/>.
		/// </summary>
		/// <param name="description">The description to which the transformed object is added.</param>
		/// <param name="solutionModifierNode">The node to transform.</param>
		/// <returns>A reference to <paramref name="description"/> (for chaining calls).</returns>
		private static SelectQueryDescription transformSolutionModifierNode(this SelectQueryDescription description, SolutionModifierNode solutionModifierNode)
        {
            if (solutionModifierNode != null)
            {
                LimitOffsetClausesNode limitOffsetClauses = solutionModifierNode.LimitOffsetClauses as LimitOffsetClausesNode;

                if (limitOffsetClauses != null)
                {
                    if (limitOffsetClauses.LimitClause != null)
                    {
                        description.Limit = (limitOffsetClauses.LimitClause as LimitClauseNode).Limit;
                    }

                    if (limitOffsetClauses.OffsetClause != null)
                    {
                        description.Offset = (limitOffsetClauses.OffsetClause as OffsetClauseNode).Offset;
                    }
                }

                if (solutionModifierNode.OrderClause != null)
                {
                    INode orderClause = solutionModifierNode.OrderClause;

                    description.OrderConditions = Util.FlattenRecursiveList<OrderCondition>(orderClause.Children[0],
                        orderClause.Children[1], orderConditionValueSelector);
                }
            }

			return description;
		}

		/// <summary>
		/// Value selector for use with <c>Util.FlattenRecursiveList</c>.
		/// </summary>
		/// <param name="node">The node.</param>
		/// <returns></returns>
		private static OrderCondition orderConditionValueSelector(INode node)
		{
			OrderCondition orderCondition = new OrderCondition();

			if (node.Children[0] is AscOrDescNode)
			{
				if ((node.Children[0] as AscOrDescNode).Order == AscOrDescNode.Orders.Ascending)
				{
					orderCondition.Type = OrderTypes.Ascending;
				}
				else
				{
					orderCondition.Type = OrderTypes.Descending;
				}

                orderCondition.Value = node.Children[1].Children[0].Children[0].Children[0].Children[0];
			}
			else
			{
                orderCondition.Value = node.Children[3];
                orderCondition.Type = OrderTypes.Ascending;
			}

			return orderCondition;
		}
	}
}
