﻿using System;
using System.Collections.Generic;
using SharQL.Ast;

namespace SharQL.Algebra
{
	/// <summary>
	/// Util class for handling recursive (head/tail) lists.
	/// </summary>
    internal static class Util
    {
		/// <summary>
		/// Flattens the recursive list.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="headNode">The head node.</param>
		/// <param name="tailNode">The tail node.</param>
		/// <param name="valueSelector">The value selector.</param>
		/// <returns>The flattened list.</returns>
		public static List<T> FlattenRecursiveList<T>(INode headNode, INode tailNode, Func<INode, T> valueSelector)
        {
            return FlattenRecursiveList<T>(headNode, tailNode, valueSelector, n => n.Children[0], n => n.Children[1]);
        }

		/// <summary>
		/// Flattens the recursive list.
		/// </summary>
		/// <typeparam name="T">The type of the list elements.</typeparam>
		/// <param name="headNode">The head of the first node.</param>
		/// <param name="tailNode">The tail of the first node.</param>
		/// <param name="valueSelector">The value selector.</param>
		/// <param name="headSelector">The head selector.</param>
		/// <param name="tailSelector">The tail selector.</param>
		/// <returns>The flattened list.</returns>
        public static List<T> FlattenRecursiveList<T>(INode headNode, INode tailNode, Func<INode, T> valueSelector, Func<INode, INode> headSelector, Func<INode, INode> tailSelector)
        {
            List<T> list = new List<T>();

			TraverseRecursiveList(list, headNode, tailNode, (n, l) => { l.Add(valueSelector(n)); }, headSelector, tailSelector);

            return list;
        }

		/// <summary>
		/// Traverses the recursive list, performing the specified action on every element.
		/// </summary>
		/// <typeparam name="T">The type of the list elements.</typeparam>
		/// <param name="argument">The object to operate on.</param>
		/// <param name="headNode">The head of the first node.</param>
		/// <param name="tailNode">The tail of the first node.</param>
		/// <param name="action">The action to perform on each node.</param>
		/// <remarks>Assumes head is first child and tail is second child.</remarks>
		public static void TraverseRecursiveList<T>(T argument, INode headNode, INode tailNode, Action<INode, T> action)
        {
			TraverseRecursiveList(argument, headNode, tailNode, action, n => n.Children[0], n => n.Children[1]);
        }

		/// <summary>
		/// Traverses the recursive list, performing the specified action on every element.
		/// </summary>
		/// <typeparam name="T">The type of the list elements.</typeparam>
		/// <param name="argument">The object to operate on.</param>
		/// <param name="headNode">The head of the first node.</param>
		/// <param name="tailNode">The tail of the first node.</param>
		/// <param name="action">The action to perform on each node.</param>
		/// <param name="headSelector">The head selector.</param>
		/// <param name="tailSelector">The tail selector.</param>
		public static void TraverseRecursiveList<T>(T argument, INode headNode, INode tailNode, Action<INode, T> action, Func<INode, INode> headSelector, Func<INode, INode> tailSelector)
        {
            INode currentHead = headNode;
            INode currentTail = tailNode;

            while (currentHead != null)
            {
                action(currentHead, argument);

                if (currentTail != null)
                {
                    currentHead = headSelector(currentTail);
                    currentTail = tailSelector(currentTail);
                }
                else
                {
                    break;
                }
            }
        }
    }
}
