﻿using System;
using System.Collections.Generic;
using System.Linq;
using SharQL.Algebra.QueryStructure;
using SharQL.Ast;
using SharQL.Ast.Nodes;

namespace SharQL.Algebra
{
	/// <summary>
	/// Contains functionality for transforming 
	/// <see cref="WhereClauseNode"/> objects into trees of 
	/// <see cref="Pattern"/>-derived objects.
	/// </summary>
	internal static class WhereClauseTransformer
	{
		/// <summary>
		/// Transforms the specified where clause into SPARQL algebra.
		/// </summary>
		/// <param name="whereClause">The where clause to transform.</param>
		/// <returns>A <see cref="Pattern"/> object representing the SPARQL 
		/// algebra expression.</returns>
		public static Pattern Transform(WhereClauseNode whereClause)
		{
			if (whereClause.Children[0] == null)
			{
				throw new ArgumentException("First child of whereClause can not be null", "whereClause");
			}

			try
			{
				return transform(whereClause.Children[0]);
			}
			catch (ArgumentException e)
			{
				throw new FormatException("Unexpected node encountered in whereClause", e);
			}
		}

		/// <summary>
		/// Transforms the specified node.
		/// </summary>
		/// <param name="node">The node to transform.</param>
		/// <returns>The resulting SPARQL algebra</returns>
		private static Pattern transform(INode node)
		{
			if (node is TriplesBlockNode)
			{
				return transformTriplesBlock((TriplesBlockNode)node);
			}
			else if (node is GroupOrUnionGraphPatternNode)
			{
				return transformGroupOrUnionGraphPattern((GroupOrUnionGraphPatternNode)node);
			}
			else if (node is GraphGraphPatternNode)
			{
				return transformGraphGraphPattern((GraphGraphPatternNode)node);
			}
			else if (node is GroupGraphPatternNode)
			{
				PatternNormalizer.NormalizeGroupGraphPattern((GroupGraphPatternNode)node);
				return transformGroupGraphPattern((GroupGraphPatternNode)node);
			}
			else
			{
				throw new ArgumentException("Unexpected node " + node.GetType().FullName, "node");
			}
		}

		/// <summary>
		/// Transforms the triples block.
		/// </summary>
		/// <param name="node">The <see cref="TriplesBlockNode"/> 
		/// node.</param>
		/// <returns>The resulting SPARQL algebra</returns>
		private static Pattern transformTriplesBlock(TriplesBlockNode node)
		{
			Triple[] triples = getListOfTriples(node);
			Pattern prevPattern = null;
			foreach (var triple in triples)
			{
				BasicGraphPattern bgp = new BasicGraphPattern();
				bgp.Triples.Add(triple);
				bgp.VariablesInScope.AddRange(getVariables(triple));

				if (prevPattern != null)
				{
					Join join = new Join(prevPattern, bgp);
					join.VariablesInScope.AddRange(prevPattern.VariablesInScope.Union(bgp.VariablesInScope));
					prevPattern = join;
				}
				else
				{
					prevPattern = bgp;
				}
			}

			return prevPattern;
		}

		/// <summary>
		/// Transforms the group or union graph pattern.
		/// </summary>
		/// <param name="node">The <see cref="GroupOrUnionGraphPatternNode"/> 
		/// node.</param>
		/// <returns>The resulting SPARQL algebra</returns>
		private static Pattern transformGroupOrUnionGraphPattern(GroupOrUnionGraphPatternNode node)
		{
			Pattern pattern = null;
			foreach (INode child in node.Children.Where(c => c != null))
			{
				if (pattern == null)
				{
					pattern = transform(child);
				}
				else
				{
					pattern = new Union(pattern, transform(child));

					pattern.VariablesInScope.AddRange(
						((Union)pattern).FirstPattern.VariablesInScope.Union(
						((Union)pattern).SecondPattern.VariablesInScope));
				}
			}
			return pattern;
		}

		/// <summary>
		/// Transforms the graph graph pattern.
		/// </summary>
		/// <param name="node">The <see cref="GraphGraphPatternNode"/> 
		/// node.</param>
		/// <returns>The resulting SPARQL algebra</returns>
		private static Pattern transformGraphGraphPattern(GraphGraphPatternNode node)
		{
			Graph graph = null;

			if (node.Children[0] is IriRefNode)
			{
				graph = new Graph((IriRefNode)node.Children[0], transform(node.Children[1]));
			}
			else if (node.Children[0] is VarNode)
			{
				graph = new Graph((VarNode)node.Children[0], transform(node.Children[1]));
			}
			else
			{
				throw new ArgumentException("Expected IriRefNode or VarNode as first child of GraphGraphPatternNode.", "node");
			}

			graph.VariablesInScope.AddRange(graph.Pattern.VariablesInScope);

			return graph;
		}

		/// <summary>
		/// Transforms the group graph pattern.
		/// </summary>
		/// <param name="node">The <see cref="GroupGraphPatternNode"/> 
		/// node.</param>
		/// <returns>The resulting SPARQL algebra.</returns>
		private static Pattern transformGroupGraphPattern(GroupGraphPatternNode node)
		{
			List<INode> filterConstraintExpressions = new List<INode>();
			Pattern pattern = new EmptyPattern();
			foreach (INode child in node.Children)
			{
				if (child == null)
				{
					continue;
				}
				else if (child is FilterNode)
				{
					filterConstraintExpressions.Add(child.Children[0]);
				}
				else if (child is OptionalGraphPatternNode)
				{
					Pattern optionalPattern = transform(child.Children[0]);
					if (optionalPattern is Filter)
					{
						pattern = new LeftJoin(pattern, ((Filter)optionalPattern).Pattern, ((Filter)optionalPattern).ConstraintExpression);
					}
					else
					{
						pattern = new LeftJoin(pattern, optionalPattern, new BooleanLiteralNode() { Value = true });
					}
					pattern.VariablesInScope.AddRange(((LeftJoin)pattern).LeftPattern.VariablesInScope
												.Union(((LeftJoin)pattern).RightPattern.VariablesInScope));
				}
				else
				{
					Pattern graphPattern = transform(child);
					if (pattern is EmptyPattern)
					{
						pattern = graphPattern;
					}
					else
					{
						pattern = new Join(pattern, graphPattern);
						pattern.VariablesInScope.AddRange(((Join)pattern).FirstPattern.VariablesInScope
													.Union(((Join)pattern).SecondPattern.VariablesInScope));
					}
				}
			}
			if (filterConstraintExpressions.Count > 0)
			{
				INode conjunctionOfConstraints = conjugateConstraints(filterConstraintExpressions);
				pattern = new Filter(conjunctionOfConstraints, pattern);
				pattern.VariablesInScope.AddRange(((Filter)pattern).Pattern.VariablesInScope);
			}
			return pattern;
		}

		/// <summary>
		/// Conjugates the specified list of constraints.
		/// </summary>
		/// <param name="filterConstraintExpressions">The filter 
		/// constraint expressions.</param>
		/// <returns>The conjugation of the specified constraints.</returns>
		private static INode conjugateConstraints(List<INode> filterConstraintExpressions)
		{
			INode conjunction = filterConstraintExpressions[0];

			for (int i = 1; i < filterConstraintExpressions.Count; i++)
			{
				conjunction = new ConditionalAndExpressionNode(conjunction as NodeBase, filterConstraintExpressions[i] as NodeBase);
			}

			return conjunction;
		}

		/// <summary>
		/// Gets the list of triples from the specified 
		/// <see cref="TriplesBlockNode"/> object.
		/// </summary>
		/// <param name="triplesBlockNodeRoot">The triples block root 
		/// node.</param>
		/// <returns>An array of <see cref="Triple"/> objects.</returns>
		private static Triple[] getListOfTriples(TriplesBlockNode triplesBlockNodeRoot)
		{
			List<Triple> triples = new List<Triple>();

			TriplesBlockNode current = triplesBlockNodeRoot;

			while (current != null)
			{
				triples.Add(translateTripleNode(current));
				current = current.Children[1] as TriplesBlockNode;
			}

			return triples.ToArray();
		}

		/// <summary>
		/// Translates a <see cref="TriplesBlockNode"/> object into a
		/// <see cref="Triple"/> object.
		/// </summary>
		/// <param name="triplesBlock">The <see cref="TriplesBlockNode"/>
		/// object to translate.</param>
		/// <returns>The resulting <see cref="Triple"/> object</returns>
		private static Triple translateTripleNode(TriplesBlockNode triplesBlock)
		{
			INode subject = null, predicate = null, @object = null;

			TriplesSameSubjectNode tssn = triplesBlock.Children[0] as TriplesSameSubjectNode;
			if (tssn != null)
			{
				subject = tssn.Children[0];

				PropertyListNode pln = tssn.Children[2] as PropertyListNode;
				if (pln != null)
				{
					predicate = pln.Children[0];

					ObjectListNode oln = pln.Children[1] as ObjectListNode;
					if (oln != null)
					{
						@object = oln.Children[0];
					}
				}
			}

			Triple triple = new Triple();

			triple.setSubject(subject);
			triple.setPredicate(predicate);
			triple.setObject(@object);

			return triple;
		}

		/// <summary>
		/// Sets the triple's subject.
		/// </summary>
		/// <param name="triple">The target triple.</param>
		/// <param name="subject">The subject to set.</param>
		private static void setSubject(this Triple triple, INode subject)
		{
			if (subject is VarNode)
			{
				triple.Subject = new Resource(ResourceType.Variable, ((VarNode)subject).Value);
			}
			else if (subject is IriRefNode)
			{
				triple.Subject = new Resource(ResourceType.Iri, ((IriRefNode)subject).Value);
			}
			else if (subject is BlankNode)
			{
				triple.Subject = new Resource(ResourceType.Blank, ((BlankNode)subject).Value);
			}
			else if (subject is RdfLiteralNode)
			{
				RdfLiteralNode node = (RdfLiteralNode)subject;

                triple.Subject = new Resource(ResourceType.Literal, '"' + ((StringLiteralNode)node.String).Value + '"');
                if (node.IriRef != null)
                {
                    triple.Subject = new Resource(triple.Subject.Type, triple.Subject.Value + "^^" + ((IriRefNode)node.IriRef).Value);
                }
                else if (!string.IsNullOrEmpty(node.LangTag))
                {
                    triple.Subject = new Resource(triple.Subject.Type, triple.Subject.Value + "@" + node.LangTag);
                }
            }
        }

		/// <summary>
		/// Sets the triple's predicate.
		/// </summary>
		/// <param name="triple">The target triple.</param>
		/// <param name="predicate">The predicate to set.</param>
		private static void setPredicate(this Triple triple, INode predicate)
		{
			if (predicate is VarNode)
			{
				triple.Predicate = new Resource(ResourceType.Variable, ((VarNode)predicate).Value);
			}
			else if (predicate is IriRefNode)
			{
				triple.Predicate = new Resource(ResourceType.Iri, ((IriRefNode)predicate).Value);
			}
		}

		/// <summary>
		/// Sets the triple's object.
		/// </summary>
		/// <param name="triple">The target triple.</param>
		/// <param name="object">The object to set.</param>
		private static void setObject(this Triple triple, INode @object)
		{
			if (@object is VarNode)
			{
				triple.Object = new Resource(ResourceType.Variable, ((VarNode)@object).Value);
			}
			else if (@object is IriRefNode)
			{
				triple.Object = new Resource(ResourceType.Iri, ((IriRefNode)@object).Value);
			}
			else if (@object is BlankNode)
			{
				triple.Object = new Resource(ResourceType.Blank, ((BlankNode)@object).Value);
			}
			else if (@object is RdfLiteralNode)
			{
				RdfLiteralNode node = (RdfLiteralNode)@object;

				triple.Object = new Resource(ResourceType.Literal, '"' + ((StringLiteralNode)((RdfLiteralNode)@object).String).Value + '"');
				if (node.IriRef != null)
				{
					triple.Object = new Resource(triple.Object.Type, triple.Object.Value + "^^" + ((IriRefNode)node.IriRef).Value);
				}
				else if (!string.IsNullOrEmpty(node.LangTag))
				{
					triple.Object = new Resource(triple.Object.Type, triple.Object.Value + "@" + node.LangTag);
				}
			}
		}

		/// <summary>
		/// Gets the variables present in the specified 
		/// <see cref="Triple"/>.
		/// </summary>
		/// <param name="triple">The triple to investigate</param>
		/// <returns>The variables present in the specified 
		/// <see cref="Triple"/>.</returns>
		private static IEnumerable<string> getVariables(Triple triple)
		{
			if (triple.Subject.Type == ResourceType.Variable)
			{
				yield return triple.Subject.Value;
			}
			else if (triple.Subject.Type == ResourceType.Blank)
			{
				yield return QueryTransformer.BlankNodeVariablePrefix + triple.Subject.Value.Substring(2);
			}

			if (triple.Predicate.Type == ResourceType.Variable)
			{
				yield return triple.Predicate.Value;
			}
			else if (triple.Predicate.Type == ResourceType.Blank)
			{
				yield return QueryTransformer.BlankNodeVariablePrefix + triple.Predicate.Value.Substring(2);
			}

			if (triple.Object.Type == ResourceType.Variable)
			{
				yield return triple.Object.Value;
			}
			else if (triple.Object.Type == ResourceType.Blank)
			{
				yield return QueryTransformer.BlankNodeVariablePrefix + triple.Object.Value.Substring(2);
			}
		}
	}
}
