﻿using System;
using System.IO;
using SharQL.Algebra.QueryStructure;

namespace SharQL.Component.Dumpers
{
	/// <summary>
	/// Formats and writes a SPARQL algebra tree to console or an arbitrary 
	/// <see cref="TextWriter"/>.
	/// </summary>
	public class AlgebraDumper
	{
		/// <summary>The <see cref="TextWriter"/> to write to.</summary>
		private TextWriter _writer;

		/// <summary>
		/// Initializes a new instance of the <see cref="AlgebraDumper"/> class.
		/// </summary>
		/// <remarks>Declared private to prevent external instantiation.</remarks>
		private AlgebraDumper()
		{ }

		/// <summary>
		/// Writes the SPARQL algebra tree below the specified root to 
		/// <see cref="Console.Out"/>.
		/// </summary>
		/// <param name="root">The root of the algebra tree.</param>
		public static void Write(Pattern root)
		{
			Write(root, Console.Out);
		}

		/// <summary>
		/// Writes the SPARQL algebra tree below the specified root to the specified 
		/// writer.
		/// </summary>
		/// <param name="root">The root of the algebra tree.</param>
		/// <param name="writer">The <see cref="TextWriter"/> to write to.</param>
		public static void Write(Pattern root, TextWriter writer)
		{
			AlgebraDumper dumper = new AlgebraDumper();
			dumper._writer = writer;
			dumper.writeObject(null, root);
		}

		/// <summary>
		/// Recursively writes the specified node with a prefix.
		/// </summary>
		/// <param name="prefix">The prefix or <c>null</c> if none.</param>
		/// <param name="node">The node to write.</param>
		private void writeObject(string prefix, Pattern node)
		{
			if (node == null) return;

			_writer.Write(prefix ?? string.Empty);
			_writer.Write(node.GetType().Name);

			if (node is BasicGraphPattern)
			{
				_writer.Write(Environment.NewLine);
				foreach (var triple in ((BasicGraphPattern)node).Triples)
				{
					writeObject((prefix ?? string.Empty) + "    ", triple);
				}
			}
			else if (node is Graph)
			{
				_writer.Write(" " + ((Graph)node).IRI != null ? ((Graph)node).IRI.Value : ((Graph)node).Variable.Value);
			
				_writer.Write(Environment.NewLine);
				writeObject((prefix ?? string.Empty) + "  ", ((Graph)node).Pattern);
			}
			else if (node is Filter)
			{
				_writer.Write(Environment.NewLine);
				writeObject((prefix ?? string.Empty) + "  ", ((Filter)node).Pattern);
			}
			else if (node is Join)
			{
				_writer.Write(Environment.NewLine);
				writeObject((prefix ?? string.Empty) + "  ", ((Join)node).FirstPattern);
				writeObject((prefix ?? string.Empty) + "  ", ((Join)node).SecondPattern);
			}
			else if (node is LeftJoin)
			{
				_writer.Write(Environment.NewLine);
				writeObject((prefix ?? string.Empty) + "  ", ((LeftJoin)node).LeftPattern);
				writeObject((prefix ?? string.Empty) + "  ", ((LeftJoin)node).RightPattern);
			}
			else if (node is Union)
			{
				_writer.Write(Environment.NewLine);
				writeObject((prefix ?? string.Empty) + "  ", ((Union)node).FirstPattern);
				writeObject((prefix ?? string.Empty) + "  ", ((Union)node).SecondPattern);
			}
		}

		/// <summary>
		/// Writes the specified triple with a prefix.
		/// </summary>
		/// <param name="prefix">The prefix.</param>
		/// <param name="triple">The triple to write.</param>
		private void writeObject(string prefix, Triple triple)
		{
			_writer.WriteLine(prefix + "{0} {1} {2}", formatNode(triple.Subject), formatNode(triple.Predicate), formatNode(triple.Object));
		}

		/// <summary>
		/// Formats the <see cref="Resource"/> object for print.
		/// </summary>
		/// <param name="resource">The resource to format.</param>
		/// <returns>A formatted string representing the resource.</returns>
		private string formatNode(Resource resource)
		{
			switch (resource.Type)
			{
				case ResourceType.Iri:
					return '<' + resource.Value + '>';

				case ResourceType.Blank:
					return resource.Value;

				case ResourceType.Variable:
					return '?' + resource.Value;
					
				case ResourceType.Literal:
				default:
					return resource.Value;
			}
		}
	}
}
