﻿using System;
using System.IO;
using SharQL.Ast;
using SharQL.Ast.Nodes;

namespace SharQL.Component.Dumpers
{
	/// <summary>
	/// Formats and writes an AST to console or an arbitrary <see cref="TextWriter"/>.
	/// </summary>
	public class AstDumper
	{
		/// <summary>The <see cref="TextWriter"/> to write to.</summary>
		private TextWriter _writer;

		/// <summary>
		/// Initializes a new instance of the <see cref="AstDumper"/> class.
		/// </summary>
		/// <remarks>Declared private to prevent external instantiation.</remarks>
		private AstDumper()
		{ }

		/// <summary>
		/// Writes the AST below the specified node to <see cref="Console.Out"/>.
		/// </summary>
		/// <param name="node">The root of the AST.</param>
		public static void Write(INode node)
		{
			Write(node, Console.Out);
		}

		/// <summary>
		/// Writes the AST below the specified node to the specified writer.
		/// </summary>
		/// <param name="node">The root of the AST.</param>
		/// <param name="writer">The <see cref="TextWriter"/> to write to.</param>
		public static void Write(INode node, TextWriter writer)
		{
			AstDumper dumper = new AstDumper();
			dumper._writer = writer;
			dumper.writeObject(null, node);
		}

		/// <summary>
		/// Recursively writes the specified node with a prefix.
		/// </summary>
		/// <param name="prefix">The prefix or <c>null</c> if none.</param>
		/// <param name="node">The node to write.</param>
		private void writeObject(string prefix, INode node)
		{
			if (node == null) return;

			_writer.Write(prefix ?? string.Empty);
			_writer.Write(node.GetType().Name);

			if (node is IriRefNode)
			{
				_writer.Write(" <" + ((IriRefNode)node).Value + ">");
			}
			else if (node is VarNode)
			{
				_writer.Write(" [" + ((VarNode)node).Value + "]");
			}
			else if (node is BlankNode)
			{
				_writer.Write(" " + ((BlankNode)node).Value);
			}

			_writer.Write(Environment.NewLine);
			foreach (INode child in node.Children)
			{
				writeObject((prefix ?? string.Empty) + "  ", child);
			}
		}
	}
}
