﻿using System.Collections.Generic;
using System.Linq;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Infrastructure;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Join;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Search;
using Microsoft.Ceres.Evaluation.Operators.Graphs;
using Microsoft.Ceres.SearchCore.Search.Operators;
using Microsoft.Ceres.SearchCore.Services.SystemCatalog;
using SharQL.Algebra;
using SharQL.Algebra.QueryStructure;
using SharQL.Ast.Nodes;

namespace SharQL.Component
{
    internal enum LookupTripleElement
    {
        None, Subject, Predicate, Object
    }

	/// <summary>
	/// Contains functionality for building Mars operator graphs from SPARQL queries.
	/// </summary>
	internal class OperatorGraphBuilder
	{
		private ISparqlQueryServiceConfiguration _config;
		private ISystemCatalog _catalog;

		/// <summary>
		/// Initializes a new instance of the <see cref="OperatorGraphBuilder"/> class.
		/// </summary>
		/// <param name="config">The configuration.</param>
		/// <param name="catalog">The system catalog.</param>
		public OperatorGraphBuilder(ISparqlQueryServiceConfiguration config, ISystemCatalog catalog)
		{
			_config = config;
			_catalog = catalog;
		}

		/// <summary>
		/// Converts the specified intermediary query representation to a Mars
		/// operator tree.
		/// </summary>
		/// <param name="query">The query to convert.</param>
		/// <returns>
		/// A tree of Mars operators, ready for evaluation.
		/// </returns>
		public OperatorGraph ConvertQueryToOperatorGraph(Query query)
		{
			OperatorBase root = new OutputOperator();
			var graph = new OperatorFlow(root);

			if (query.Description is SelectQueryDescription)
			{
				SelectQueryDescription desc = (SelectQueryDescription)query.Description;
				InsertWhereClause(graph, root, desc.WhereClause);
                InsertOrderBy(graph, root, desc.OrderConditions);
                InsertProjection(graph, root, desc.Variables, desc.WhereClause.VariablesInScope);
                InsertDistinctReduced(graph, root, desc.DistinctReducedModifier, desc.Variables, desc.WhereClause.VariablesInScope);
                InsertLimitOffset(graph, root, desc.Limit, desc.Offset);
			}
			else
			{
				// TODO: Support other query types
			}

			return graph;
		}

		/// <summary>
		/// Inserts operators for a distinct/reduced modifier. If <paramref name="selectedVariables"/>
		/// variables is non-empty, these variable are used for comparison; otherwise, the variables of
		/// <paramref name="allVariables"/> are used.
		/// </summary>
		/// <param name="graph">The graph to add operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="modifier">The distinct/reduced modifier.</param>
		/// <param name="selectedVariables">Any selected variables.</param>
		/// <param name="allVariables">All variables of the query.</param>
		private void InsertDistinctReduced(OperatorFlow graph, OperatorBase successor, DistinctReducedModifiers modifier, IEnumerable<string> selectedVariables, IEnumerable<string> allVariables)
		{
			if (modifier == DistinctReducedModifiers.Distinct)
			{
				SortOperator sortOperator = new SortOperator();

				var variableSource = OperatorGraphUtils.SelectFirstNonemptyList(selectedVariables, allVariables);
				sortOperator.AddSortFields(variableSource);

				ProjectDistinctOperator projectDistinctOperator = new ProjectDistinctOperator { Prefix = variableSource.Count() };

				graph.InsertBefore(successor, projectDistinctOperator);
				graph.InsertBefore(projectDistinctOperator, sortOperator);
			}
			else if (modifier == DistinctReducedModifiers.Reduced)
			{
				// TODO: Not explicitly handled in prototype
			}
		}

		/// <summary>
		/// Inserts an operator to achieve ordering.
		/// </summary>
		/// <param name="graph">The graph to add the operator to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="orderConditions">The order conditions.</param>
		private void InsertOrderBy(OperatorFlow graph, OperatorBase successor, IEnumerable<OrderCondition> orderConditions)
		{
			if (orderConditions != null && orderConditions.Count() > 0)
			{
				SortOperator sortOperator = new SortOperator();
				foreach (var condition in orderConditions)
				{
					// TODO: Support expressions
					if (condition.Value is VarNode)
					{
						sortOperator.AddSortField(((VarNode)condition.Value).Value, condition.Type == OrderTypes.Ascending);
					}
				}

				graph.InsertBefore(successor, sortOperator);
			}
		}

		/// <summary>
		/// Inserts an operator to achieve a projection. If <paramref name="selectedVariables"/>
		/// variables is non-empty, these variable are projected; otherwise, the variables of
		/// <paramref name="allVariables"/> are projected.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="selectedVariables">Any selected variables.</param>
		/// <param name="allVariables">All variables of the query.</param>
		private void InsertProjection(OperatorFlow graph, OperatorBase successor, IEnumerable<string> selectedVariables, IEnumerable<string> allVariables)
		{
			MapOperator mapOperator = new MapOperator(false) { ParameterMap = new Dictionary<IdentifierProperty, ExpressionProperty>() };
			var variableSource = OperatorGraphUtils.SelectFirstNonemptyList(selectedVariables, allVariables);

            // Ignore blank node surrogate variables
            variableSource = variableSource.Where(s => !s.StartsWith(QueryTransformer.BlankNodeVariablePrefix));
			
            mapOperator.ParameterMap.AddVariables(variableSource);

			graph.InsertBefore(successor, mapOperator);
		}

		/// <summary>
		/// Inserts an operator to achieve any offset or limiting of the result set.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="limit">The limit specifier, if any.</param>
		/// <param name="offset">The offset specifier, if any.</param>
		private void InsertLimitOffset(OperatorFlow graph, OperatorBase successor, int? limit, int? offset)
		{
			if (limit.HasValue || offset.HasValue)
			{
				var trimOp = new TrimOperator();
				if (limit.HasValue)
				{
					trimOp.Hits = limit.Value;
				}
				if (offset.HasValue)
				{
					trimOp.Offset = offset.Value;
				}
				graph.InsertBefore(successor, trimOp);
			}
		}

		/// <summary>
		/// Inserts the operators necessary to achieve the results specified
		/// by the where clause.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="whereClauseAlgebra">The where clause algebra.</param>
		private void InsertWhereClause(OperatorFlow graph, OperatorBase successor, Pattern whereClauseAlgebra)
		{
			AddTranslatedAlgebra(graph, successor, whereClauseAlgebra);
		}

		/// <summary>
		/// Translates the algebra of <paramref name="whereClauseAlgebra"/> into Mars operators
		/// and adds it to the specified graph before the specified successor operator.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="whereClauseAlgebra">The where clause algebra to translate.</param>
		private void AddTranslatedAlgebra(OperatorFlow graph, OperatorBase successor, Pattern whereClauseAlgebra)
		{
			if (whereClauseAlgebra is BasicGraphPattern)
			{
				AddBasicGraphPattern(graph, successor, (BasicGraphPattern)whereClauseAlgebra);
			}
			else if (whereClauseAlgebra is Filter)
			{
				AddFilter(graph, successor, (Filter)whereClauseAlgebra);
			}
			else if (whereClauseAlgebra is Graph)
			{
				AddGraphSpecification(graph, successor, (Graph)whereClauseAlgebra);
			}
			else if (whereClauseAlgebra is Join)
			{
				AddJoin(graph, successor, (Join)whereClauseAlgebra);
			}
			else if (whereClauseAlgebra is LeftJoin)
			{
				AddLeftJoin(graph, successor, (LeftJoin)whereClauseAlgebra);
			}
			else if (whereClauseAlgebra is Union)
			{
				AddUnion(graph, successor, (Union)whereClauseAlgebra);
			}
		}

		/// <summary>
		/// Adds an operator to achieve the specified union.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="union">The union to achieve.</param>
		private void AddUnion(OperatorFlow graph, OperatorBase successor, Union union)
		{
			SortOperator leftSortOperator = new SortOperator();
			SortOperator rightSortOperator = new SortOperator();

			foreach (var variable in union.FirstPattern.VariablesInScope)
			{
				leftSortOperator.AddSortField(variable, true);
			}

			foreach (var variable in union.SecondPattern.VariablesInScope)
			{
				rightSortOperator.AddSortField(variable, true);
			}

			OrOperator orOperator = new OrOperator();

			graph.AddPredecessor(successor, orOperator);
			graph.AddPredecessor(orOperator, leftSortOperator);
			graph.AddPredecessor(orOperator, rightSortOperator);

			AddTranslatedAlgebra(graph, leftSortOperator, union.FirstPattern);
			AddTranslatedAlgebra(graph, rightSortOperator, union.SecondPattern);

			// ----- Code below does not work as expected -----

			//UnionOperator unionOperator = new UnionOperator();
			//unionOperator.UnionType = UnionType.All;

			//graph.AddPredecessor(successor, unionOperator);
			//AddTranslatedAlgebra(graph, unionOperator, union.FirstPattern);
			//AddTranslatedAlgebra(graph, unionOperator, union.SecondPattern);
		}

		/// <summary>
		/// Adds an operator to achieve the specified left join.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="leftJoin">The left join to achieve.</param>
		private void AddLeftJoin(OperatorFlow graph, OperatorBase successor, LeftJoin leftJoin)
		{
            // TODO: Not planned for protoype
		}

		/// <summary>
		/// Adds operators to achieve the specified join.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="join">The join to achieve.</param>
		/// <remarks>A join between two pattern which have no variables in common results
		/// in a Cartesian product, rather than a natural join.</remarks>
		private void AddJoin(OperatorFlow graph, OperatorBase successor, Join join)
		{
			// Remapping removes prefixes added by MergeJoin
			MapOperator remapOperator = CreateJoinMapping(join);
			graph.AddPredecessor(successor, remapOperator);

			if (HasCommonVariables(join.FirstPattern, join.SecondPattern))
			{
				AddNaturalJoin(graph, remapOperator, join);
			}
			else
			{
				AddCartesianProduct(graph, remapOperator, join);
			}
		}

		/// <summary>
		/// Adds operators to achieve a natural join.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="join">The join to achieve.</param>
		private void AddNaturalJoin(OperatorFlow graph, OperatorBase successor, Join join)
		{
			OperatorBase joinOperator = null;
			MapOperator mapOperator1 = null, mapOperator2 = null;
			SortOperator sortOperator1 = null, sortOperator2 = null;

			CreateNaturalJoin(join, out joinOperator, out mapOperator1, out mapOperator2,
				out sortOperator1, out sortOperator2);

			graph.AddPredecessor(successor, joinOperator);

			graph.AddPredecessor(joinOperator, sortOperator1);
			graph.AddPredecessor(joinOperator, sortOperator2);

			graph.AddPredecessor(sortOperator1, mapOperator1);
			graph.AddPredecessor(sortOperator2, mapOperator2);

			AddTranslatedAlgebra(graph, mapOperator1, join.FirstPattern);
			AddTranslatedAlgebra(graph, mapOperator2, join.SecondPattern);
		}

		/// <summary>
		/// Adds operators to achieve a cartesian product.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="join">The join to achieve.</param>
		private void AddCartesianProduct(OperatorFlow graph, OperatorBase successor, Join join)
		{
			OperatorBase joinOperator = null;

			CreateCartesianProduct(join, out joinOperator);

			graph.AddPredecessor(successor, joinOperator);

			AddTranslatedAlgebra(graph, joinOperator, join.FirstPattern);
			AddTranslatedAlgebra(graph, joinOperator, join.SecondPattern);
		}

		/// <summary>
		/// Adds operators to achieve the specified graph specification.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="graphSpecification">The graph specification to achieve.</param>
		/// <remarks>Not implemented for prototype.</remarks>
		private void AddGraphSpecification(OperatorFlow graph, OperatorBase successor, Graph graphSpecification)
		{
			// TODO: Not planned for protoype
		}

		/// <summary>
		/// Adds operators to achieve the specified filter.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="filter">The filter to achieve.</param>
		/// <remarks>Not implemented for prototype.</remarks>
		private void AddFilter(OperatorFlow graph, OperatorBase successor, Filter filter)
		{
			// TODO: Not planned for protoype
		}

		/// <summary>
		/// Adds operators to query the index according to the specified basic
		/// graph pattern.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="bgp">The basic graph pattern to query for.</param>
		private void AddBasicGraphPattern(OperatorFlow graph, OperatorBase successor, BasicGraphPattern bgp)
		{
			foreach (var triple in bgp.Triples)
			{
				AddLookup(graph, successor, triple);
			}
		}

		/// <summary>
		/// Adds operators to query the index according to the specified triple pattern.
		/// </summary>
		/// <param name="graph">The graph to add the operators to.</param>
		/// <param name="successor">The successor operator.</param>
		/// <param name="triple">The triple pattern to query for.</param>
		private void AddLookup(OperatorFlow graph, OperatorBase successor, Triple triple)
		{
			LookupOperator lookupOperator = null;
			SelectOperator selectOperator = null;
			MapOperator mapOperator = null;

			LookupTripleElement lookupTripleElement = CreateLookupAndSelect(triple, out lookupOperator, out selectOperator);
			CreateLookupVariableMapping(triple, lookupTripleElement, out mapOperator);

			if (lookupOperator != null)
			{
				graph.AddPredecessor(successor, mapOperator);
				successor = mapOperator;

				if (selectOperator != null)
				{
					graph.AddPredecessor(successor, selectOperator);
					successor = selectOperator;
				}
				graph.AddPredecessor(successor, lookupOperator);
			}
		}

		/// <summary>
		/// Creates the mapping from index fields to variables, according to the specified
		/// triple pattern.
		/// </summary>
		/// <param name="triple">The triple pattern.</param>
		/// <param name="mapOperator">[Out] The resulting map operator.</param>
		private void CreateLookupVariableMapping(Triple triple, LookupTripleElement lookupTripleElement, out MapOperator mapOperator)
		{
			mapOperator = new MapOperator(false) { ParameterMap = new Dictionary<IdentifierProperty, ExpressionProperty>() };

			if (triple.Subject.Type == ResourceType.Variable)
			{
				mapOperator.ParameterMap[new IdentifierProperty(triple.Subject.Value)] = new IdentifierProperty(_config.FirstValueField);
			}
            else if (triple.Subject.Type == ResourceType.Blank)
            {
                mapOperator.ParameterMap[new IdentifierProperty(QueryTransformer.BlankNodeVariablePrefix + triple.Subject.Value.Substring(2))] = new IdentifierProperty(_config.FirstValueField);
            }

			if (triple.Predicate.Type == ResourceType.Variable)
			{
				mapOperator.ParameterMap[new IdentifierProperty(triple.Predicate.Value)] =
						new IdentifierProperty(lookupTripleElement == LookupTripleElement.Subject ?
								_config.FirstValueField : _config.SecondValueField
						);
			}
            else if (triple.Predicate.Type == ResourceType.Blank)
            {
                mapOperator.ParameterMap[new IdentifierProperty(QueryTransformer.BlankNodeVariablePrefix + triple.Predicate.Value.Substring(2))] =
                        new IdentifierProperty(lookupTripleElement == LookupTripleElement.Subject ?
                                _config.FirstValueField : _config.SecondValueField
                        );
            }

			if (triple.Object.Type == ResourceType.Variable)
			{
				mapOperator.ParameterMap[new IdentifierProperty(triple.Object.Value)] = new IdentifierProperty(_config.SecondValueField);
			}
            else if (triple.Object.Type == ResourceType.Blank)
            {
                mapOperator.ParameterMap[new IdentifierProperty(QueryTransformer.BlankNodeVariablePrefix + triple.Object.Value.Substring(2))] = new IdentifierProperty(_config.SecondValueField);
            }
		}

		/// <summary>
		/// Creates the lookup and select operators that query and filter the index, according
		/// to the specified triple pattern.
		/// </summary>
		/// <param name="triple">The triple pattern.</param>
		/// <param name="lookupOperator">[Out] The resulting lookup operator.</param>
		/// <param name="selectOperator">[Out] The resulting select operator.</param>
		private LookupTripleElement CreateLookupAndSelect(Triple triple, out LookupOperator lookupOperator, out SelectOperator selectOperator)
		{
			lookupOperator = null;
			selectOperator = null;

			if (triple.Object.Type != ResourceType.Variable && triple.Object.Type != ResourceType.Blank)
			{
				lookupOperator = new LookupOperator(_catalog)
				{
					IndexName = _config.OccurrenceIndex,
					Word = _config.ObjectPrefix + triple.Object.Value
				};

				selectOperator = CreateSelectFilter(triple.Subject, triple.Predicate, null);
                return LookupTripleElement.Object;
			}
			else if (triple.Subject.Type != ResourceType.Variable && triple.Subject.Type != ResourceType.Blank)
			{
				lookupOperator = new LookupOperator(_catalog)
				{
					IndexName = _config.OccurrenceIndex,
					Word = _config.SubjectPrefix + triple.Subject.Value
				};

				selectOperator = CreateSelectFilter(null, triple.Predicate, triple.Object);
                return LookupTripleElement.Subject;
			}
			else if (triple.Predicate.Type != ResourceType.Variable && triple.Predicate.Type != ResourceType.Blank)
			{
				lookupOperator = new LookupOperator(_catalog)
				{
					IndexName = _config.OccurrenceIndex,
					Word = _config.PredicatePrefix + triple.Predicate.Value
				};

				selectOperator = CreateSelectFilter(triple.Subject, null, triple.Object);
                return LookupTripleElement.Predicate;
			}
            else
            {
                return LookupTripleElement.None;
            }
		}

		/// <summary>
		/// Creates the cartesian product operator that fulfils the specified join specification.
		/// </summary>
		/// <param name="join">The join specification.</param>
		/// <param name="joinOperator">[Out] The resulting join operator.</param>
		private void CreateCartesianProduct(
			Join join,
			out OperatorBase joinOperator)
		{
			joinOperator = new MergeJoinOperator { JoinPrefix = 0 };
		}

		/// <summary>
		/// Creates the natural join operator that fulfils the specified join specification.
		/// </summary>
		/// <param name="join">The join specification.</param>
		/// <param name="joinOperator">[Out] The resulting join operator.</param>
		/// <param name="mapOperator1">[Out] The first resulting map operator.</param>
		/// <param name="mapOperator2">[Out] The second resulting map operator.</param>
		/// <param name="sortOperator1">[Out] The first resulting sort operator.</param>
		/// <param name="sortOperator2">[Out] The second resulting sort operator.</param>
		private void CreateNaturalJoin(
			Join join,
			out OperatorBase joinOperator,
			out MapOperator mapOperator1,
			out MapOperator mapOperator2,
			out SortOperator sortOperator1,
			out SortOperator sortOperator2)
		{
			var commonVariables = GetCommonVariables(join.FirstPattern, join.SecondPattern);

			mapOperator1 = new MapOperator(false) { ParameterMap = new Dictionary<IdentifierProperty, ExpressionProperty>() };
			mapOperator2 = new MapOperator(false) { ParameterMap = new Dictionary<IdentifierProperty, ExpressionProperty>() };
			joinOperator = new MergeJoinOperator { JoinPrefix = commonVariables.Count() };
			sortOperator1 = new SortOperator();
			sortOperator2 = new SortOperator();

			mapOperator1.ParameterMap
				.AddVariables(commonVariables)
				.AddVariables(join.FirstPattern.VariablesInScope.Except(commonVariables));

			mapOperator2.ParameterMap
				.AddVariables(commonVariables)
				.AddVariables(join.SecondPattern.VariablesInScope.Except(commonVariables));

			foreach (var variable in commonVariables)
			{
				sortOperator1.AddSortField(variable, true);
				sortOperator2.AddSortField(variable, true);
			}
		}

		/// <summary>
		/// Creates the map operator that removes prefix artifacts left by merge join.
		/// </summary>
		/// <param name="join">The join specification.</param>
		/// <returns>The resulting map operator.</returns>
		private MapOperator CreateJoinMapping(Join join)
		{
			MapOperator remapOperator = new MapOperator(false) { ParameterMap = new Dictionary<IdentifierProperty, ExpressionProperty>() };

			var commonVariables = GetCommonVariables(join.FirstPattern, join.SecondPattern);

			remapOperator.ParameterMap
				.AddVariables(commonVariables)
				.AddVariables(join.FirstPattern.VariablesInScope.Except(commonVariables), string.Empty, "Input0.")
				.AddVariables(join.SecondPattern.VariablesInScope.Except(commonVariables), string.Empty, "Input1.");

			return remapOperator;
		}

		/// <summary>
		/// Creates the operator that filter output according to the specified resources.
		/// One of the resource parameters is assumed to be <c>null</c>, indicating that
		/// it is constant after being used in a lookup.
		/// </summary>
		/// <param name="subjectResource">The subject resource.</param>
		/// <param name="predicateResource">The predicate resource.</param>
		/// <param name="objectResource">The object resource.</param>
		/// <returns>The resulting select operator if filtering is meaningful; otherwise
		/// <c>null</c>.</returns>
		private SelectOperator CreateSelectFilter(Resource subjectResource, Resource predicateResource, Resource objectResource)
		{
			string filter = CreateTripleFilter(subjectResource, predicateResource, objectResource);

			if (filter != null)
			{
				SelectOperator selectOperator = new SelectOperator
				{
					Filter = new ExpressionProperty(filter),
				};

				return selectOperator;
			}
			else
			{
				return null;
			}
		}

        private bool IsVariableOrBlankResource(Resource resource)
        {
            return (resource.Type == ResourceType.Variable || resource.Type == ResourceType.Blank);
        }

        private bool IsBlankResource(Resource resource)
        {
            return resource != null && resource.Type == ResourceType.Blank;
        }

		/// <summary>
		/// Creates the filter expression according to the specified resources. One of the
		/// resource parameters is assumed to be <c>null</c>, indicating that it is constant
		/// after being used in a lookup.
		/// </summary>
		/// <param name="subjectFilter">The subject filter.</param>
		/// <param name="predicateFilter">The predicate filter.</param>
		/// <param name="objectFilter">The object filter.</param>
		/// <returns>The resulting filter expression if filtering is meaningful; otherwise
		/// <c>null</c>.</returns>
		private string CreateTripleFilter(Resource subjectFilter, Resource predicateFilter, Resource objectFilter)
		{
			var predicates = new List<string>();

			if (subjectFilter != null && !IsVariableOrBlankResource(subjectFilter))
			{
				predicates.Add(string.Format("({0} == StringFunctions.ToSafeString(\"{1}\"))", _config.FirstValueField, subjectFilter.Value.Replace("\"", "\\\"")));
			}
            else if (IsBlankResource(subjectFilter))
            {
                predicates.Add(string.Format("(Filters.StartsWith({0}, \"_:\"))", _config.FirstValueField));
            }

            if (predicateFilter != null && !IsVariableOrBlankResource(predicateFilter))
			{
				predicates.Add(string.Format("({0} == StringFunctions.ToSafeString(\"{1}\"))", subjectFilter == null ?
					_config.FirstValueField : _config.SecondValueField,
					predicateFilter.Value.Replace("\"", "\\\"")));
			}
            else if (IsBlankResource(predicateFilter))
            {
                predicates.Add(string.Format("(Filters.StartsWith({0}, \"_:\"))", subjectFilter == null ?
                    _config.FirstValueField : _config.SecondValueField));
            }

            if (objectFilter != null && !IsVariableOrBlankResource(objectFilter))
			{
				predicates.Add(string.Format("({0} == StringFunctions.ToSafeString(\"{1}\"))", _config.SecondValueField, objectFilter.Value.Replace("\"", "\\\"")));
			}
            else if (IsBlankResource(objectFilter))
            {
                predicates.Add(string.Format("(Filters.StartsWith({0}, \"_:\"))", _config.SecondValueField));
            }

			if (predicates.Count > 0)
			{
				return string.Join(" && ", predicates.ToArray());
			}
			else
			{
				return null;
			}
		}

		/// <summary>
		/// Determines whether the two specified patterns have any variables in common.
		/// </summary>
		/// <param name="pattern1">The first pattern.</param>
		/// <param name="pattern2">The second pattern.</param>
		/// <returns>
		/// <c>true</c> if the two specified patterns have any variables in common; 
		/// otherwise, <c>false</c>.
		/// </returns>
		private bool HasCommonVariables(Pattern pattern1, Pattern pattern2)
		{
			return GetCommonVariables(pattern1, pattern2).Count() > 0;
		}

		/// <summary>
		/// Gets the variables which the two specified patterns have in common.
		/// </summary>
		/// <param name="pattern1">The first pattern.</param>
		/// <param name="pattern2">The second pattern.</param>
		/// <returns>The variables which the two specified patterns have in common.</returns>
		private IEnumerable<string> GetCommonVariables(Pattern pattern1, Pattern pattern2)
		{
			return pattern1.VariablesInScope.Intersect(pattern2.VariablesInScope);
		}
	}
}
