﻿using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.BuiltIn.Basic;
using Microsoft.Ceres.Evaluation.Operators.Graphs;

namespace SharQL.Component
{
	/// <summary>
	/// Static class containing utility methods for handling operator graphs.
	/// </summary>
	internal static class OperatorGraphUtils
	{
		/// <summary>
		/// Selects the first nonempty list.
		/// </summary>
		/// <typeparam name="T">The type of the list elements</typeparam>
		/// <param name="alternatives">The alternatives.</param>
		/// <returns>The first nonempty list.</returns>
		public static IEnumerable<T> SelectFirstNonemptyList<T>(params IEnumerable<T>[] alternatives)
		{
			return alternatives.First(list => list.Count() > 0);
		}

		/// <summary>
		/// Gets validation error messages.
		/// </summary>
		/// <param name="graph">The graph.</param>
		/// <returns>A string describing any validation errors.</returns>
		public static string GetValidationErrors(OperatorGraph graph)
		{
			var sb = new StringBuilder("Validation error(s):\n");

			foreach (var statusMessage in graph.GetOperatorStatusMessages())
			{
				sb.AppendLine(statusMessage.Reporter + ": " + statusMessage.Message);
			}

			return sb.ToString();
		}

		/// <summary>
		/// Adds variables to a parameter map, using the specified prefixes for keys and values.
		/// </summary>
		/// <param name="parameterMap">The parameter map.</param>
		/// <param name="variables">The variables to add.</param>
		/// <param name="keyPrefix">The prefix to prepend keys with.</param>
		/// <param name="valuePrefix">The prefix to prepend values with.</param>
		/// <returns>Reference to self (for call chaining).</returns>
		public static IDictionary<IdentifierProperty, ExpressionProperty> AddVariables(
			this IDictionary<IdentifierProperty, ExpressionProperty> parameterMap,
			IEnumerable<string> variables, string keyPrefix, string valuePrefix)
		{
			foreach (var variable in variables)
			{
				parameterMap[new IdentifierProperty(keyPrefix + variable)] = new IdentifierProperty(valuePrefix + variable);
			}
			return parameterMap;
		}

		/// <summary>
		/// Adds variables to a parameter map.
		/// </summary>
		/// <param name="parameterMap">The parameter map.</param>
		/// <param name="variables">The variables to add.</param>
		/// <returns>Reference to self (for call chaining).</returns>
		public static IDictionary<IdentifierProperty, ExpressionProperty> AddVariables(
			this IDictionary<IdentifierProperty, ExpressionProperty> parameterMap,
			IEnumerable<string> variables)
		{
			return AddVariables(parameterMap, variables, string.Empty, string.Empty);
		}

		/// <summary>
		/// Adds the sort fields.
		/// </summary>
		/// <param name="sortOperator">The sort operator.</param>
		/// <param name="fields">The fields to add.</param>
		/// <param name="naturalOrder">if set to <c>true</c> natural order (ascending) is used.</param>
		/// <returns>Reference to self (for call chaining).</returns>
		public static SortOperator AddSortFields(this SortOperator sortOperator, IEnumerable<string> fields, bool naturalOrder)
		{
			foreach (var field in fields)
			{
				sortOperator.AddSortField(field, naturalOrder);
			}
			return sortOperator;
		}

		/// <summary>
		/// Adds the sort fields, assuming natural ordering (ascending).
		/// </summary>
		/// <param name="sortOperator">The sort operator.</param>
		/// <param name="fields">The fields to add.</param>
		/// <returns>Reference to self (for call chaining).</returns>
		public static SortOperator AddSortFields(this SortOperator sortOperator, IEnumerable<string> fields)
		{
			return AddSortFields(sortOperator, fields, true);
		}
	}
}
