﻿using System;
using System.Collections.Generic;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.Graphs;
using Microsoft.Ceres.Evaluation.Services;
using Microsoft.Ceres.SearchCore.Services.QueryEngine;
using Microsoft.Ceres.SearchCore.Services.SystemCatalog;
using SharQL.Algebra;
using SharQL.Algebra.QueryStructure;
using SharQL.Ast;
using SharQL.Ast.Nodes;

namespace SharQL.Component
{
    /// <summary>
    /// Service for handling SPARQL queries. Provides parsing as well as conversion
    /// to operator trees.
    /// </summary>
    public class SparqlQueryService : ISparqlQueryService
    {
        private Parser _parser;
		private IQueryEngine _queryEngine;
		private OperatorGraphBuilder _graphBuilder;

		/// <summary>
		/// Gets the graph builder.
		/// </summary>
		/// <value>The graph builder.</value>
		/// <remarks>Only included for access by interactive shell.</remarks>
		internal OperatorGraphBuilder GraphBuilder
		{
			get { return _graphBuilder; }
			set { _graphBuilder = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="SparqlQueryService"/> class.
		/// </summary>
		/// <param name="queryEngine">The query engine.</param>
		/// <param name="config">The configuration.</param>
		/// <param name="catalog">The system catalog.</param>
		public SparqlQueryService(IQueryEngine queryEngine, ISparqlQueryServiceConfiguration config, ISystemCatalog catalog)
        {
            _parser = new SharQL.Parser();
			_queryEngine = queryEngine;
			_graphBuilder = new OperatorGraphBuilder(config, catalog);
        }

        #region ISparqlQueryService Members

		/// <summary>
		/// Executes the given SPARQL query.
		/// </summary>
		/// <param name="query">The SPARQL query to be executed.</param>
		/// <returns>XML-formatted result set.</returns>
        public string ExecuteQuery(string query)
        {
            var generator = new OutputXmlGenerator();

            try
            {
                DoExecuteQuery(query, generator);
                return generator.ToString();
            }
            catch (Exception e)
            {
                generator.Failure("Caught exception: " + e);
                return generator.ToString();
            }
        }

		/// <summary>
		/// A list of errors found while parsing the last query.
		/// </summary>
        public IEnumerable<string> Errors
        {
            get
            {
                foreach (var error in _parser.Errors)
                {
                    yield return string.Format("{0} at line {1}, column {2}, position {3}.", 
						error.message, 
						error.span.sLin.ToString(),
                        error.span.sCol.ToString(), 
						error.span.sPos.ToString());
                }
            }
        }

        #endregion

		/// <summary>
		/// Executes the specified query and writes the results to the specified
		/// generator.
		/// </summary>
		/// <param name="query">The query to execute.</param>
		/// <param name="generator">The generator to write results to.</param>
        internal void DoExecuteQuery(string query, OutputXmlGenerator generator)
        {
			if (_queryEngine == null)
            {
                generator.Failure("No query engine was injected.");
            }
            else if (query == null || query.Trim().Length == 0)
            {
                generator.Failure("Empty query.");
            }
			else if (_parser.Parse(query))
            {
                OperatorGraph operatorGraph = _graphBuilder.ConvertQueryToOperatorGraph(ConvertAstToQuery(_parser.AstRoot));
				ValidateAndExecute(operatorGraph, generator);
            }
            else
            {
                foreach (var error in Errors)
                {
                    generator.Failure(error);
                }
            }
        }

		/// <summary>
		/// Validates and executes the specified operator graph, printing the results
		/// to the specified generator.
		/// </summary>
		/// <param name="operatorGraph">The operator graph to execute.</param>
		/// <param name="generator">The generator to write results to.</param>
		private void ValidateAndExecute(OperatorGraph operatorGraph, OutputXmlGenerator generator)
		{
			ValidationStatus validationStatus = operatorGraph.ValidateOperators();

			if (validationStatus == ValidationStatus.Ok)
			{
				// Execute
				using (QueryContext queryContext = _queryEngine.ExecuteQuery(operatorGraph))
				{
					IFlowHandle flowHandle = null;

					if (queryContext != null)
					{
						flowHandle = queryContext.Result;
					}
					else
					{
						generator.Failure("Unable to obtain query context");
					}

					if (flowHandle != null)
					{
						var results = flowHandle.OutputRecordSets;
						generator.FetchMainResult(results);
					}
					else
					{
						generator.Failure("Problem during executing query. Received null flow handle or local handle from Query Engine.");
					}
				}
			}
			else
			{
				generator.Failure(OperatorGraphUtils.GetValidationErrors(operatorGraph));
			}
		}

        /// <summary>
        /// Parses the specified source.
        /// </summary>
        /// <param name="source">The source to parse.</param>
        /// <returns>The root of the resulting AST.</returns>
        public INode Parse(string source)
        {
            if (_parser.Parse(source))
            {
                return _parser.AstRoot;
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Converts the specified AST root to a <see cref="Query"/> object.
        /// </summary>
        /// <param name="astRoot">The AST root to convert.</param>
        /// <returns>
        /// An intermediary representation of the SPARQL query.
        /// </returns>
        public Query ConvertAstToQuery(INode astRoot)
        {
            return QueryTransformer.Transform(astRoot as QueryNode);
        }
    }
}
