﻿using System;
using System.IO;
using Microsoft.Ceres.CoreServices.Services.Container;
using Microsoft.Ceres.CoreServices.Services.DependencyInjection;
using Microsoft.Ceres.CoreServices.Services.Node;
using Microsoft.Ceres.SearchCore.Services.QueryEngine;
using Microsoft.Ceres.SearchCore.Services.SystemCatalog;
using Microsoft.Ceres.Evaluation.Operators.Expressions;
using Microsoft.Ceres.Evaluation.Operators.Expressions.FunctionPlugIn;
using Microsoft.Ceres.Common.Utils;

namespace SharQL.Component
{
    public static class Filters
    {
        [FunctionInfo("")]
        public static bool? StartsWith(object arg, string prefix)
        {
            return (bool?)arg.ToString().StartsWith(prefix);
        }
    }

	/// <summary>
	/// The Mars component encapsulating an <see cref="ISparqlQueryService"/>.
	/// </summary>
	[ContainerManaged]
	public class SparqlQueryServiceComponent : AbstractManageableCoreClient, IInteractiveComponent
	{
		private SparqlQueryService _queryService;
		private ISparqlQueryServiceConfiguration _configuration;

		/// <summary>
		/// Initializes a new instance of the <see cref="SparqlQueryServiceComponent"/> class.
		/// </summary>
		public SparqlQueryServiceComponent()
		{ 
            FunctionRegistry.Instance.AddType(typeof(Filters));
        }

		/// <summary>
		/// Gets or sets the query engine.
		/// </summary>
		[Injected]
		public IQueryEngine QueryEngine { get; set; }

		/// <summary>
		/// Gets or sets the system catalog.
		/// </summary>
		[Injected(Optional = false)]
		public ISystemCatalog Catalog { get; set; }

		/// <summary>
		/// Gets the SPARQL query service.
		/// </summary>
		/// <value>The SPARQL query service.</value>
		[Exposed(ExposureState = ComponentState.Active)]
		public ISparqlQueryService QueryService
		{
			get { return _queryService; }
		}

		/// <summary>
		/// Gets the SPARQL query service as a <see cref="SparqlQueryService"/>
		/// object.
		/// </summary>
		/// <value>The SPARQL query service as a <see cref="SparqlQueryService"/>
		/// object.</value>
		internal SparqlQueryService QueryServiceImpl
		{
			get { return _queryService; }
		}

		/// <summary>
		/// Gets the type of the configuration.
		/// </summary>
		/// <value>The type of the configuration.</value>
		public override Type ConfigurationType
		{
			get { return typeof(ISparqlQueryServiceConfiguration); }
		}

		/// <summary>
		/// Occurs when a new configuration is applied.
		/// </summary>
		/// <param name="previousConfiguration">The previous configuration.</param>
		/// <param name="newConfiguration">The new configuration.</param>
		protected override void Configured(object previousConfiguration, object newConfiguration)
		{
			ISparqlQueryServiceConfiguration configuration = newConfiguration as ISparqlQueryServiceConfiguration;
			if (configuration != null)
			{
				_configuration = configuration;
			}
		}

		/// <summary>
		/// Occurs when all dependencies are resolved and the component is being 
		/// activated.
		/// </summary>
		public override void Activating()
		{
			base.Activating();
			_queryService = new SparqlQueryService(QueryEngine, _configuration, Catalog);
		}

		#region IInteractiveComponent Members

		/// <summary>
		/// Starts interactive mode for the component.
		/// </summary>
		/// <param name="reader">The reader to read from.</param>
		/// <param name="writer">The writer to write to.</param>
		public void Interactive(TextReader reader, TextWriter writer)
		{
			SparqlQueryServiceShell shell = new SparqlQueryServiceShell(this, reader, writer);
			shell.Interactive();
		}

		#endregion
	}
}
