﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.Ceres.CoreServices.InteractiveShell;
using Microsoft.Ceres.Evaluation.Operators;
using Microsoft.Ceres.Evaluation.Operators.Graphs;
using SharQL.Algebra.QueryStructure;
using SharQL.Ast;
using SharQL.Component.Dumpers;

namespace SharQL.Component
{
	/// <summary>
	/// Handles shell interaction on behalf of a <see cref="SparqlQueryServiceComponent"/>.
	/// </summary>
	public class SparqlQueryServiceShell : AbstractInteractiveShell
	{
		/// <summary>The component represented.</summary>
		private SparqlQueryServiceComponent _component;

		/// <summary>Test query for usage when no query is provided.</summary>
		private const string testQuery = @"PREFIX : <http://example.org/> SELECT * WHERE { { ?s :p1 ""literal""@en-GB } UNION {?s :p2 3.6 } UNION { ?s :p3 2 } } ORDER BY ?s OFFSET 3 LIMIT 5";

		/// <summary>
		/// Initializes a new instance of the <see cref="SparqlQueryServiceShell"/> class.
		/// </summary>
		/// <param name="component">The component to represent.</param>
		/// <remarks>Reads from <see cref="Console.In"/>; writes to <see cref="Console.Out"/></remarks>
		public SparqlQueryServiceShell(SparqlQueryServiceComponent component)
			:this(component, Console.In, Console.Out)
		{}

		/// <summary>
		/// Initializes a new instance of the <see cref="SparqlQueryServiceShell"/> class.
		/// </summary>
		/// <param name="component">The component to represent.</param>
		/// <param name="reader">The reader to read from.</param>
		/// <param name="writer">The writer to write to.</param>
		public SparqlQueryServiceShell(SparqlQueryServiceComponent component, TextReader reader, TextWriter writer)
			:base(reader, writer)
		{
			Prompt = "Sparql";
			_component = component;
		}

		/// <summary>
		/// Parses input and writes AST.
		/// </summary>
		/// <param name="args">The query to parse.</param>
		[Command(
			Command = "ParseToAst",
			Description = "Parses a SPARQL query to AST representation",
			Help = "Parses a SPARQL query and shows its AST representation.\n")]
		public void ParseToAst(string[] args)
		{
			StringBuilder output = new StringBuilder();
			string input = string.Join(" ", args);

			if (string.IsNullOrEmpty(input.Trim()))
			{
				input = testQuery;
				WriteLine("No query provided; using test query:");
				WriteLine(testQuery);
				WriteLine();
			}

			try
			{
				AstDumper.Write(_component.QueryServiceImpl.Parse(input), new StringWriter(output));

				WriteLine("\n" + output.ToString());
			}
			catch (Exception e)
			{
				WriteLine("ERROR:");
				Write(e.StackTrace);
				WriteLine();
			}
		}

		/// <summary>
		/// Parses input and writes SPARQL algebra.
		/// </summary>
		/// <param name="args">The query to parse.</param>
		[Command(
			Command = "ParseToAlgebra",
			Description = "Parses a SPARQL query to algebraic representation",
			Help = "Parses a SPARQL query and shows its algebraic representation.\n")]
		public void ParseToAlgebra(string[] args)
		{
			StringBuilder output = new StringBuilder();
			string input = string.Join(" ", args);

			if (string.IsNullOrEmpty(input.Trim()))
			{
				input = testQuery;
				WriteLine("No query provided; using test query:");
				WriteLine(testQuery);
				WriteLine();
			}

			try
			{
				INode root = _component.QueryServiceImpl.Parse(input);
				Query query = _component.QueryServiceImpl.ConvertAstToQuery(root);

				AlgebraDumper.Write(query.Description.WhereClause, new StringWriter(output));

				WriteLine("\n" + output.ToString());
			}
			catch(Exception e)
			{
				WriteLine("ERROR:");
				Write(e.StackTrace);
				WriteLine();
			}
		}

		/// <summary>
		/// Parses input and writes operator tree.
		/// </summary>
		/// <param name="args">The query to parse.</param>
        [Command(
            Command = "ParseToOperators",
            Description = "Parses a SPARQL query to Mars operators",
            Help = "Parses a SPARQL query and shows the resulting operator tree.\n")]
        public void ParseToOperators(string[] args)
		{
			//const string testQuery = @"SELECT DISTINCT * WHERE {{?a ?b 'c' . ?a ?a <http://www.example.com/>} UNION {?a ?b 'c'}} ORDER BY ?a OFFSET 3 LIMIT 5";

            StringBuilder output = new StringBuilder();
			string input = string.Join(" ", args);

			if (string.IsNullOrEmpty(input.Trim()))
			{
				input = testQuery;
				WriteLine("No query provided; using test query:");
				WriteLine(testQuery);
				WriteLine();
			}

            INode root = _component.QueryServiceImpl.Parse(input.ToString());

            if (root != null)
            {
                Query query = _component.QueryServiceImpl.ConvertAstToQuery(root);
                OperatorGraph ops = _component.QueryServiceImpl.GraphBuilder.ConvertQueryToOperatorGraph(query);

                if (ops.ValidateOperators() == ValidationStatus.Ok)
                {
                    OperatorDumper.Write(ops.Roots.First(), new StringWriter(output));
                    WriteLine("\n" + output.ToString());
                }
                else
                {
                    WriteLine("Operator graph validation failed!");
                }
            }
            else
            {
                foreach (var error in _component.QueryService.Errors)
                {
                    WriteLine(error);
                }
            }
        }

		/// <summary>
		/// Executes a SPARQL query contained in the specified file.
		/// </summary>
		/// <param name="args">The file containing the query to execute.</param>
        [Command(
            Command = "FileExecuteQuery",
			Description = "Executes a SPARQL query from file",
			Help = "Executes a SPARQL query from file and shows the results.\n")]
        public void FileExecuteQuery(string[] args)
        {
			if (_component.QueryService != null)
			{
				string file = string.Join(" ", args);
				if (!File.Exists(file))
				{
					WriteLine("The file could not be found.");
				}
				else
				{
					string query;
					using (FileStream stream = new FileStream(file, FileMode.Open))
					using (StreamReader reader = new StreamReader(stream))
					{
						query = reader.ReadToEnd();
					}
					WriteLine("Executing query:\n" + query + "\n");
					WriteLine(_component.QueryService.ExecuteQuery(query));
				}
			}
			else
			{
				WriteLine("Dependencies are not available. Please try again later.");
			}
        }

		/// <summary>
		/// Executes a SPARQL query.
		/// </summary>
		/// <param name="args">The query to execute.</param>
		[Command(
			Command = "ExecuteQuery",
			Description = "Executes a SPARQL query",
			Help = "Executes a SPARQL query and shows the results.\n")]
		public void ExecuteQuery(string[] args)
		{
			if (_component.QueryService != null)
			{
				WriteLine(_component.QueryService.ExecuteQuery(string.Join(" ", args)));
			}
			else
			{
				WriteLine("Dependencies are not available. Please try again later.");
			}
		}
	}
}
