﻿using System;
using System.Collections.Generic;
using SharQL.Ast.Visitor;

namespace SharQL.Ast
{

	/// <summary>
	/// The base implementation of the <see cref="INode"/> interface.
	/// </summary>
	public abstract class NodeBase : INode
	{
		#region Protected Fields

		/// <summary>The parent of the node.</summary>
		protected INode parent;
		/// <summary>The children of the node.</summary>
		protected IList<INode> children;

		#endregion

		/// <summary>
		/// Default constructor.
		/// </summary>
		public NodeBase()
		{
			children = new NodeCollection();
		}

		/// <summary>
		/// Constructs the object and adds a set of children.
		/// </summary>
		/// <param name="children">The children to add</param>
		public NodeBase(params INode[] children)
			: this()
		{
			foreach (INode node in children)
			{
				if (node != null) node.Parent = this;
				this.children.Add(node);
			}
		}

		/// <summary>
		/// Gets or sets the parent of the node. Should be set to
		/// <c>null</c> if node is root.
		/// </summary>
		public virtual INode Parent
		{
			get { return parent; }
			set { parent = value; }
		}

		/// <summary>
		/// Gets the list of children of the node.
		/// </summary>
		public virtual IList<INode> Children
		{
			get { return children; }
		}

		/// <summary>
		/// Accepts a visitor. The visitor pattern can be used for
		/// e.g. optimizing the AST.
		/// </summary>
		/// <param name="visitor">The visitor</param>
		public virtual void Accept(IVisitor visitor)
		{
			Accept(visitor, VisitorOrder.ChildrenBeforeParent);
		}

		/// <summary>
		/// Accepts a visitor. The visitor pattern can be used for
		/// e.g. optimizing the AST.
		/// </summary>
		/// <param name="visitor">The visitor</param>
		/// <param name="order">The visitor order</param>
		public virtual void Accept(IVisitor visitor, VisitorOrder order)
		{
			if (order == VisitorOrder.ChildrenBeforeParent)
			{
				for (int i = children.Count - 1; i >= 0; i--)
				{
					INode child = children[i];

					if (child != null)
						child.Accept(visitor, order);
				}
				Action<INode> action = visitor[GetType()];
				if (action != null)
				{
					action(this);
				}
			}
			else if (order == VisitorOrder.ParentBeforeChildren)
			{
				Action<INode> action = visitor[GetType()];
				if (action != null)
				{
					action(this);
				}
				for (int i = children.Count - 1; i >= 0; i--)
				{
					INode child = children[i];

					if (child != null)
						child.Accept(visitor, order);
				}
			}
		}

		/// <summary>
		/// Returns a string representation of the node.
		/// </summary>
		/// <returns>A string representation of the node</returns>
		public override string ToString()
		{
			return this.GetType().FullName;
		}

		/// <summary>
		/// Returns a prefixed string representation of the node.
		/// </summary>
		/// <param name="prefix">The string prefix to which the string
		/// representation should be appended.</param>
		/// <returns>A prefixed string representation of the node</returns>
		public virtual string ToString(string prefix)
		{
			return prefix + ToString();
		}

		#region ICloneable Members

		/// <summary>
		/// Creates a new <see cref="NodeBase"/> object, with it memebers
		/// initialized to the same values as the current object.
		/// </summary>
		/// <returns>a new <see cref="NodeBase"/> object, with it memebers
		/// initialized to the same values as the current object</returns>
		public virtual NodeBase Clone()
		{
			NodeBase newNode = (NodeBase)MemberwiseClone();
			newNode.children = new NodeCollection();
			foreach (INode node in children)
			{
				if (node != null)
				{
					newNode.children.Add(node.Clone() as INode);
				}
				else
				{
					newNode.children.Add(null);
				}
			}

			return newNode;
		}

		object ICloneable.Clone()
		{
			return Clone();
		}

		#endregion
	}
}
