﻿using System;
using System.Collections.ObjectModel;
using MPLEX.Parser;
using SharQL.Ast;
using SharQL.Ast.Visitor;
using SharQL.Utils;

namespace SharQL
{
	/// <summary>
	/// Creates abstrast syntax tree representations from SPARQL queries.
	/// </summary>
	public class Parser
	{
		#region Private Fields

		private Scanner scanner;
		private ParsingEngine parser;
        private SharQLErrorHandler errorHandler;

		#endregion

        /// <summary>
        /// Gets the errors from the last parsing as a read-only collection.
        /// </summary>
        public ReadOnlyCollection<Error> Errors
        {
            get { return errorHandler.SortedErrorList().AsReadOnly(); }
        }

        /// <summary>
		/// Gets the root node of the abstract syntax tree.
		/// </summary>
		/// <remarks>If no parsing has been performed, or if previous parsing
		/// attempt failed, the returned value will be <c>null</c>.</remarks>
		public INode AstRoot { get { return parser.AstRoot; } }

		/// <summary>
		/// Creates a new <see cref="Parser"/> object.
		/// </summary>
		public Parser()
		{
			parser = new ParsingEngine();
		}

		/// <summary>
		/// Attempts to parse the specified SPARQL source.
		/// </summary>
		/// <param name="source">The SPARQL source</param>
		/// <param name="offset">The offset (in characters) to where the
		/// SPARQL query begins</param>
		/// <returns><c>true</c> if successful, <c>false</c> otherwise</returns>
		public bool Parse(string source, int offset)
		{
			string unescapedSource = string.Empty;
			
			scanner = new Scanner();
			errorHandler = new SharQLErrorHandler();

			try
			{
				unescapedSource = Utils.EscapeSequenceResolver.ResolveCodepointEscapeSequences(source);
			}
			catch (Exception e)
			{
				errorHandler.AddError(e.Message, new LexSpan());
				return false;
			}

            scanner.SetSource(unescapedSource, offset);
			scanner.Handler = errorHandler;
			
			parser.scanner = scanner;
			parser.AstRoot = null;

            if (parser.Parse())
            {
                // Post-processing
                try
                {
                    parser.AstRoot.Accept(new EscapeSequenceResolverVisitor());
                }
                catch (Exception e)
                {
					errorHandler.AddError(e.Message, new LexSpan());
                    return false;
                }
                return true;
            }
            else
            {
                return false;
            }
		}

		/// <summary>
		/// Attempts to parse the specified SPARQL source.
		/// </summary>
		/// <param name="source">The SPARQL source</param>
		/// <returns><c>true</c> if successful, <c>false</c> otherwise</returns>
		public bool Parse(string source)
		{
			return Parse(source, 0);
		}
	}
}
