%using SharQL.Ast.Nodes
%using SharQL.Ast
%using SharQL.Utils
%using System.Globalization
%using System.Diagnostics

/* Defines the generated parser namespace. */
%namespace SharQL

/* Makes the generated parser class a partial class. */
%partial

/* Defines the generated parser class name. */
%parsertype ParsingEngine

/* Defines the token location object type name. */
%YYLTYPE CustomLexLocation

%{

// Provides access to the generated Scanner class instance.
private Scanner sc { get { return scanner as Scanner; } }

%}

/*
 * SPARQL Query Language for RDF
 * W3C Recommendation 15 January 2008
 * 
 * http://www.w3.org/TR/2008/REC-rdf-sparql-query-20080115/ 
 *
 */

%start Query

%token IRI_REF PNAME_NS PNAME_LN BLANK_NODE_LABEL VAR1 VAR2 LANGTAG 
%token INTEGER DECIMAL DOUBLE INTEGER_POSITIVE DECIMAL_POSITIVE 
%token DOUBLE_POSITIVE INTEGER_NEGATIVE DECIMAL_NEGATIVE 
%token DOUBLE_NEGATIVE 
%token EXPONENT STRING_LITERAL1 STRING_LITERAL2 STRING_LITERAL_LONG1 
%token STRING_LITERAL_LONG2 ECHAR NIL ANON  
%token BASE PREFIX SELECT DISTINCT REDUCED CONSTRUCT DESCRIBE ASK FROM 
%token NAMED WHERE ORDERBY ASC DESC LIMIT OFFSET OPTIONAL 
%token GRAPH FILTER 
%token UNION OR AND NOT LESSOREQUAL GREATEROREQUAL STR LANG LANGMATCHES 
%token DATATYPE BOUND SAMETERM ISIRI ISURI ISBLANK 
%token ISLITERAL REGEX TRUE 
%token FALSE HATS

%union { public NodeBase Value; }

%%

/* Bracketed numbers refer to SPARQL grammar rule numbers.  */

/* [1] */
Query
: Prologue SelectQuery
{
	this.AstRoot = new QueryNode($1, $2);
}
| Prologue ConstructQuery 
{
	this.AstRoot = new QueryNode($1, $2);
}
| Prologue DescribeQuery
{
	this.AstRoot = new QueryNode($1, $2);
}
| Prologue AskQuery
{
	this.AstRoot = new QueryNode($1, $2);
}
;

/* [2] */
Prologue
/* REDUCE/REDUCE
: BaseDecl
{
	$$ = new PrologueNode($1, null);
} */
: BaseDecl PrefixDeclList
{
	$$ = new PrologueNode($1, $2);
}
| PrefixDeclList
{
	$$ = new PrologueNode(null, $1);
}
/* 
| REDUCE/REDUCE
{
	$$ = null;
	Debug.WriteLine("No Prologue");
}*/
;

PrefixDeclList
: PrefixDecl PrefixDeclList
{
	$$ = new PrefixDeclListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of PrefixDeclList");
}
;

/* [3] */
BaseDecl
: BASE IriRef
{
	$$ = new PrefixDeclNode($2);
}
;

/* [4] */
PrefixDecl
: PREFIX Pname IriRef
{
	PrefixDeclNode pn = (PrefixDeclNode)$2;
	$$ = new PrefixDeclNode(pn.Namespace, $3);
}
;

Pname
: PNAME_NS
{
	$$ = new PrefixDeclNode(sc.yytext, null);
}
;

IriRef
: IRI_REF
{
	$$ = new IriRefNode() { Value = sc.yytext };
}
;

/* [5] */
SelectQuery
: SELECT DistinctReducedModifier StarVarList DatasetClauseList WhereClause SolutionModifier
{
	$$ = new SelectQueryNode($2, $3, $4, $5, $6);
}
;

DistinctReducedModifier
: DISTINCT
{
	$$ = new DistinctReducedModifierNode()
	{
		Modifier = DistinctReducedModifierNode.Modifiers.Distinct
	};
}
| REDUCED
{
	$$ = new DistinctReducedModifierNode()
	{
		Modifier = DistinctReducedModifierNode.Modifiers.Reduced
	};
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("No DistinctReducedModified");
}
;

StarVarList
: '*'
{
	$$ = new StarVarListNode();
}
| Var VarList
{
	$$ = new StarVarListNode($1, $2);
}
;

VarList
: Var VarList
{
	$$ = new VarListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty VarList");
}
;

DatasetClauseList
: DatasetClause DatasetClauseList
{
	$$ = new DatasetClauseListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty DataSetClauseList");
}
;

/* [6] */
ConstructQuery
: CONSTRUCT ConstructTemplate DatasetClauseList	WhereClause SolutionModifier
{
	$$ = new ConstructQueryNode($2, $3, $4, $5);
}
;

/* [7] */
DescribeQuery
: DESCRIBE StarVarOrIRIrefList DatasetClauseList ConditionalWhereClause SolutionModifier
{
	$$ = new DescribeQueryNode($2, $3, $4, $5);
}
;

StarVarOrIRIrefList
: '*'
{
	$$ = new StarVarOrIRIrefListNode();
}
| VarOrIRIref VarOrIRIrefList
{
	$$ = new StarVarOrIRIrefListNode($1, $2);
}
;

VarOrIRIrefList
: VarOrIRIref VarOrIRIrefList
{
	$$ = new VarOrIRIrefListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty VarOrIRIrefList");
}
;

ConditionalWhereClause
: WhereClause
{
	$$ = new ConditionalWhereClauseNode($1);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty ConditionalWhereClause");
}
;

/* [8] */
AskQuery
: ASK DatasetClauseList WhereClause
{
	$$ = new AskQueryNode($2, $3);
}
;

/* [9] */
DatasetClause
: FROM DefaultOrNamedGraphClause
{
	$$ = new DatasetClauseNode($2);
}
;

DefaultOrNamedGraphClause
: DefaultGraphClause
{
	$$ = new DefaultOrNamedGraphClauseNode($1)
	{
		Type = DefaultOrNamedGraphClauseNode.Types.Default
	};
}
| NamedGraphClause
{
	$$ = new DefaultOrNamedGraphClauseNode($1)
	{
		Type = DefaultOrNamedGraphClauseNode.Types.Named
	};
}
;	

/* [10] */
DefaultGraphClause
: SourceSelector
{
	$$ = new DefaultGraphClauseNode($1);
}
;

/* [11] */
NamedGraphClause
: NAMED SourceSelector
{
	$$ = new NamedGraphClauseNode($2);
}
;

/* [12] */
SourceSelector
: IRIref
{
	$$ = new SourceSelectorNode($1);
}
;

/* [13] */
WhereClause
: WHERE GroupGraphPattern
{
	$$ = new WhereClauseNode($2);
}
| GroupGraphPattern 
{
	$$ = new WhereClauseNode($1);
}
;

/* [14] */
SolutionModifier
: OrderClause
{
	$$ = new SolutionModifierNode($1, null); 
}
| LimitOffsetClauses
{
	$$ = new SolutionModifierNode(null, $1); 
}
| OrderClause LimitOffsetClauses
{
	$$ = new SolutionModifierNode($1, $2); 
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty SolutionModifier");
}
;

/* [15] */
LimitOffsetClauses
: LimitClause
{
	$$ = new LimitOffsetClausesNode($1, null);
}
| LimitClause OffsetClause
{
	$$ = new LimitOffsetClausesNode($1, $2);
}
| OffsetClause
{
	$$ = new LimitOffsetClausesNode(null, $1);
}
| OffsetClause LimitClause
{
	$$ = new LimitOffsetClausesNode($2, $1);
}
;

/* [16] */
OrderClause
: ORDERBY OrderCondition OrderConditionList
{
	$$ = new OrderClauseNode($2, $3);
}
;

OrderConditionList
: OrderCondition OrderConditionList
{
	$$ = new OrderConditionListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("Empty OrderConditionList");
}
;

/* [17] */
OrderCondition
: AscOrDesc BrackettedExpression
{
	$$ = new OrderConditionNode($1, $2, null, null);
}
| Constraint
{
	$$ = new OrderConditionNode(null, null, $1, null);
}
| Var
{
	$$ = new OrderConditionNode(null, null, null, $1);
}
;

AscOrDesc
: ASC
{
	$$ = new AscOrDescNode() 
	{ 
		Order = AscOrDescNode.Orders.Ascending
	};
}
| DESC
{
	$$ = new AscOrDescNode() 
	{ 
		Order = AscOrDescNode.Orders.Descending
	};
}
;

/* [18] */
LimitClause
: LIMIT INTEGER
{
	$$ = new LimitClauseNode()
	{
		Limit = Convert.ToInt32(sc.yytext)
	};
}
;

/* [19] */
OffsetClause
: OFFSET INTEGER
{
	$$ = new OffsetClauseNode()
	{
		Offset = Convert.ToInt32(sc.yytext)
	};
}
;

/* [20] */
GroupGraphPattern
: '{' ConditionalTriplesBlock GroupGraphPatternBlockList '}'
{
	$$ = new GroupGraphPatternNode($2, $3);
}
;

ConditionalTriplesBlock
: TriplesBlock
{
	$$ = $1;
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("No TriplesBlock");
}
;

GroupGraphPatternBlockList
: GroupGraphPatternBlock GroupGraphPatternBlockList
{
	$$ = new GroupGraphPatternBlockListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of GroupGraphPatternBlockList");
}
;

GroupGraphPatternBlock
: GraphPatternNotTriples ConditionalDot ConditionalTriplesBlock
{
	$$ = new GroupGraphPatternBlockNode($1, null, $3);
}
| Filter ConditionalDot ConditionalTriplesBlock
{
	$$ = new GroupGraphPatternBlockNode(null, $1, $3);
}
;

ConditionalDot
: '.'
{
	Debug.WriteLine("Dot");
}
| /* empty */
{
	Debug.WriteLine("No Dot");
}
;

/* [21] */
TriplesBlock
: TriplesSameSubject ConditionalDottedTriplesBlock
{
	$$ = new TriplesBlockNode($1, $2);
}
;

ConditionalDottedTriplesBlock
: '.'
{
	$$ = null;
}
| '.' TriplesBlock
{
	$$ = $2;
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("No ConditionalDottedTriplesBlock");
}
;

/* [22] */
GraphPatternNotTriples
: OptionalGraphPattern 
{
	$$ = $1;
}
| GroupOrUnionGraphPattern 
{
	$$ = $1;
}
| GraphGraphPattern
{
	$$ = $1;
}
;

/* [23] */
OptionalGraphPattern
: OPTIONAL GroupGraphPattern
{
	$$ = new OptionalGraphPatternNode($2);
}
;

/* [24] */
GraphGraphPattern
: GRAPH VarOrIRIref GroupGraphPattern
{
	$$ = new GraphGraphPatternNode($1, $2);
}
;

/* [25] */
GroupOrUnionGraphPattern
: GroupGraphPattern GroupGraphPatternUnionList
{
	$$ = new GroupOrUnionGraphPatternNode($1, $2);
}
;

GroupGraphPatternUnionList
: UNION GroupGraphPattern GroupGraphPatternUnionList
{
	$$ = new GroupGraphPatternUnionListNode($2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of GroupGraphPatternUnionList");
}
;

/* [26] */
Filter
: FILTER Constraint
{
	$$ = new FilterNode($2);
}
;

/* [27] */
Constraint
: BrackettedExpression 
{
	$$ = $1;
}
| BuiltInCall 
{
	$$ = $1;
}
| FunctionCall
{
	$$ = $1;
}
;

/* [28] */
FunctionCall
: IRIref ArgList
{
	$$ = new FunctionCallNode($1, $2);
}
;

/* [29] */
ArgList
: NIL
{
	$$ = new ArgListNode();
}
| '(' Expression ExpressionList ')'
{
	$$ = new ArgListNode($2, $3);
}
;

ExpressionList
: ',' Expression ExpressionList
{
	$$ = new ExpressionListNode($2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of ExpressionList");
}
;

/* [30] */
ConstructTemplate
: '{' ConditionalConstructTriples '}'
{
	$$ = $2;
}
;

ConditionalConstructTriples
: ConstructTriples
{
	$$ = $1;
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("No ConstructTriples");
}
;

/* [31]  */
ConstructTriples
: TriplesSameSubject ConditionalConstructDottedTriples
{
	$$ = new ConstructTriplesNode($1, $2);
}
;

ConditionalConstructDottedTriples
: '.'
{
	$$ = new ConditionalConstructDottedTriplesNode(null);
}
| '.' ConstructTriples
{
	$$ = new ConditionalConstructDottedTriplesNode($2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("No ConstructTriples");
}
;

/* [32] */
TriplesSameSubject
: VarOrTerm PropertyListNotEmpty
{
	$$ = new TriplesSameSubjectNode($1, null, $2);
}
| TriplesNode PropertyList
{
	$$ = new TriplesSameSubjectNode(null, $1, $2);
}
;

/* [33] */
PropertyListNotEmpty
/* REDUCE/REDUCE
: Verb ObjectList 
{
	$$ = new PropertyListNode($1, $2, null);
}*/
: Verb ObjectList PropertyListTail
{
	$$ = new PropertyListNode($1, $2, $3);
}
;

PropertyListTail
: ';' PropertyListTail
{
	$$ = $2;
}
| ';' Verb ObjectList PropertyListTail
{
	$$ = new PropertyListNode($2, $3, $4);
}
| /* empty */ 
{
	$$ = null;
	Debug.WriteLine("End of PropertyList");
}
;

/* [34] */
PropertyList
: PropertyListNotEmpty
{
	$$ = $1;
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of PropertyList");
}
;

/* [35] */
ObjectList
/* REDUCE/REDUCE
: Object
{
	$$ = new ObjectListNode($1, null);
} */
: Object ObjectListTail
{
	$$ = new ObjectListNode($1, $2);
}
;

ObjectListTail
: ',' Object ObjectListTail
{
	$$ = new ObjectListNode($2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of ObjectListTail");
}
;

/* [36] */
Object
: GraphNode
{
	$$ = $1;
}
;

/* [37] */
Verb
: VarOrIRIref  
{
	$$ = $1;
}
| 'a'
{
	$$ = new IriRefNode()
	{
		Value = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
	};
}
;

/* [38] */
TriplesNode
: Collection 
{
	$$ = $1;
}
| BlankNodePropertyList
{
	$$ = $1;
}
;

/* [39] */
BlankNodePropertyList
: '[' PropertyListNotEmpty ']'
{
	$$ = new BlankNodePropertyListNode($2);
}
;

/* [40] */
Collection
: '(' GraphNode GraphNodeList ')'
{
	$$ = new CollectionNode(new GraphNodeListNode($2, $3));
}
;

GraphNodeList
: GraphNode GraphNodeList
{
	$$ = new GraphNodeListNode($1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of GraphNodeList");
}
;

/* [41] */
GraphNode
: VarOrTerm
{
	$$ = $1;
}
| TriplesNode
{
	$$ = $1;
}
;

/* [42] */
VarOrTerm
: Var 
{
	$$ = $1;
}
| GraphTerm
{
	$$ = $1;
}
;

/* [43] */
VarOrIRIref
: Var
{
	$$ = $1;
}
| IRIref
{
	$$ = $1;
}
;

/* [44] */
Var
: VAR1
{
	$$ = new VarNode(sc.yytext.Substring(1));
}
| VAR2
{
	$$ = new VarNode(sc.yytext.Substring(1));
}
;

/* [45] */
GraphTerm
: IRIref
{
	$$ = $1;
}
| RDFLiteral
{
	$$ = $1;
}
| NumericLiteral
{
	$$ = $1;
}
| BooleanLiteral
{
	$$ = $1;
}
| BlankNode
{
	$$ = $1;
}
| NIL
{
	$$ = null;
}
;

/* [46] */
Expression
: ConditionalOrExpression
{
	$$ = $1;
}
;

/* [47] */
ConditionalOrExpression
: ConditionalAndExpression ConditionalOrExpressionList
{
	$$ = new ConditionalOrExpressionNode($1, $2);
}
;

ConditionalOrExpressionList
: OR ConditionalAndExpression ConditionalOrExpressionList
{
	$$ = new ConditionalOrExpressionListNode($2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of ConditionalOrExpressionList");
}
;

/* [48] */
ConditionalAndExpression
: ValueLogical ConditionalAndExpressionList
{
	$$ = new ConditionalAndExpressionNode($1, $2);
}
;

ConditionalAndExpressionList
: AND ValueLogical ConditionalAndExpressionList
{
	$$ = new ConditionalAndExpressionListNode($2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of ConditionalAndExpressionList");
}
;

/* [49] */
ValueLogical
: RelationalExpression
{
	$$ = $1;
}
;

/* [50] */
RelationalExpression
: NumericExpression
{
	$$ = $1;
}
| NumericExpression '=' NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.Equal, $1, $3);
}
| NumericExpression NOT NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.Not, $1, $3);
}
| NumericExpression '<' NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.Less, $1, $3);
}
| NumericExpression '>' NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.Greater, $1, $3);
}
| NumericExpression LESSOREQUAL NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.LessOrEqual, $1, $3);
}
| NumericExpression GREATEROREQUAL NumericExpression
{
	$$ = new RelationalExpressionNode(Constants.Operators.GreaterOrEqual, $1, $3);
}
;

/* [51] */
NumericExpression
: AdditiveExpression
{
	$$ = $1;
}
;

/* [52] */
AdditiveExpression
: MultiplicativeExpression AdditiveExpressionList
{
	$$ = new AdditiveExpressionNode($1, $2);
}
;

AdditiveExpressionList
: '+' MultiplicativeExpression AdditiveExpressionList
{
	$$ = new AdditiveExpressionListNode(Constants.Operators.Add, $1, $2);
}
| '-' MultiplicativeExpression AdditiveExpressionList
{
	$$ = new AdditiveExpressionListNode(Constants.Operators.Subtract, $1, $2);
}
| NumericLiteralPositive AdditiveExpressionList
{
	$$ = new AdditiveExpressionListNode(Constants.Operators.None, $1, $2);
}
| NumericLiteralNegative AdditiveExpressionList 
{
	$$ = new AdditiveExpressionListNode(Constants.Operators.None, $1, $2);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of AdditiveExpressionList");
}
;

/* [53] */
MultiplicativeExpression
: UnaryExpression UnaryExpressionList
{
	$$ = new MultiplicativeExpressionNode($1, $2);
}
;

UnaryExpressionList
: '*' UnaryExpression UnaryExpressionList
{
	$$ = new UnaryExpressionListNode(Constants.Operators.Multiply, $2, $3);
}
| '/' UnaryExpression UnaryExpressionList
{
	$$ = new UnaryExpressionListNode(Constants.Operators.Divide, $2, $3);
}
| /* empty */
{
	$$ = null;
	Debug.WriteLine("End of UnaryExpressionList");
}
;

/* [54] */
UnaryExpression
: '!' PrimaryExpression
{
	$$ = new UnaryExpressionNode(Constants.Operators.Not, $2);
}
| '+' PrimaryExpression
{
	$$ = new UnaryExpressionNode(Constants.Operators.Add, $2);
}
| '-' PrimaryExpression
{
	$$ = new UnaryExpressionNode(Constants.Operators.Subtract, $2);
}
| PrimaryExpression
{
	$$ = $1;
}
;

/* [55] */
PrimaryExpression
: BrackettedExpression
{
	$$ = $1;
}
| BuiltInCall
{
	$$ = $1;
}
| IRIrefOrFunction
{
	$$ = $1;
}
| RDFLiteral 
{
	$$ = $1;
}
| NumericLiteral
{
	$$ = $1;
}
| BooleanLiteral 
{
	$$ = $1;
}
| Var
{
	$$ = $1;
}
;

/* [56] */
BrackettedExpression
: '(' Expression ')'
{
	$$ = $2;
}
;

/* [57] */
BuiltInCall
: STR '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.Str, $3);
}
| LANG '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.Lang, $3);
}
| LANGMATCHES '(' Expression ',' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.LangMatches, $3, $5);
}
| DATATYPE '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.Datatype, $3);
}
| BOUND '(' Var ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.Bound, $3);
}
| SAMETERM '(' Expression ',' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.SameTerm, $3, $5);
}
| ISIRI '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.IsIri, $3);
}
| ISURI '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.IsUri, $3);
}
| ISBLANK '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.IsBlank, $3);
}
| ISLITERAL '(' Expression ')' 
{
	$$ = new BuiltInCallNode(BuiltInCallNode.CallType.IsLiteral, $3);
}
| RegexExpression
{
	$$ = $1;
}
;

/* [58] */
RegexExpression
: REGEX '(' Expression ',' Expression ')'
{
	$$ = new RegexExpressionNode($3, $5);
}
| REGEX '(' Expression ',' Expression ',' Expression ')'
{
	$$ = new RegexExpressionNode($3, $5, $7);
}
;

/* [59] */
IRIrefOrFunction
: IRIref
{
	$$ = $1;
}
| IRIref ArgList
{
	$$ = new FunctionCallNode($1, $2);
}
;

/* [60] */
RDFLiteral
: String 
{
	$$ = new RdfLiteralNode($1, null);
}
| String LANGTAG 
{
	$$ = new RdfLiteralNode($1, null) { LangTag = sc.yytext };
}
| String HATS IRIref
{
	$$ = new RdfLiteralNode($1, $3);
}
;

/* [61] */
NumericLiteral
: NumericLiteralUnsigned
{
	$$ = $1;
} 
| NumericLiteralPositive
{
	$$ = $1;
} 
| NumericLiteralNegative
{
	$$ = $1;
}
;

/* [62] */
NumericLiteralUnsigned
: INTEGER 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Int, 
		IntValue = Convert.ToInt32(sc.yytext) };
}
| DECIMAL 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Decimal, 
		DecimalValue = Convert.ToDecimal(sc.yytext, CultureInfo.InvariantCulture) };
}
| DOUBLE
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Double, 
		DoubleValue = Convert.ToDouble(sc.yytext, CultureInfo.InvariantCulture) };
}
;

/* [63] */
NumericLiteralPositive
: INTEGER_POSITIVE 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Int, 
		IntValue = Convert.ToInt32(sc.yytext) };
}
| DECIMAL_POSITIVE 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Decimal, 
		DecimalValue = Convert.ToDecimal(sc.yytext, CultureInfo.InvariantCulture) };
}
| DOUBLE_POSITIVE
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Double, 
		DoubleValue = Convert.ToDouble(sc.yytext, CultureInfo.InvariantCulture) };
}
;

/* [64] */
NumericLiteralNegative
: INTEGER_NEGATIVE 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Int, 
		IntValue = Convert.ToInt32(sc.yytext) };
}
| DECIMAL_NEGATIVE 
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Decimal, 
		DecimalValue = Convert.ToDecimal(sc.yytext, CultureInfo.InvariantCulture) };
}
| DOUBLE_NEGATIVE
{
	$$ = new NumericLiteralNode() { Type = NumericLiteralNode.LiteralType.Double, 
		DoubleValue = Convert.ToDouble(sc.yytext, CultureInfo.InvariantCulture) };
}
;

/* [65] */
BooleanLiteral
: TRUE
{
	$$ = new BooleanLiteralNode() { Value = true };
}
| FALSE
{
	$$ = new BooleanLiteralNode() { Value = false };
}
;

/* [66] */
String
: STRING_LITERAL1 
{
	string str = sc.yytext;
	$$ = new StringLiteralNode() { Value = str.Substring(1, str.Length - 2) };
}
| STRING_LITERAL2 
{
	string str = sc.yytext;
	$$ = new StringLiteralNode() { Value = str.Substring(1, str.Length - 2) };
}
| STRING_LITERAL_LONG1 
{
	string str = sc.yytext;
	$$ = new StringLiteralNode() { Value = str.Substring(3, str.Length - 6) };
}
| STRING_LITERAL_LONG2
{
	string str = sc.yytext;
	$$ = new StringLiteralNode() { Value = str.Substring(3, str.Length - 6) };
}
;

/* [67] */
IRIref
: IRI_REF
{
	$$ = new IriRefNode() { Value = sc.yytext };
}
| PrefixedName
{
	$$ = $1;
}
;

/* [68] */
PrefixedName
: PNAME_LN 
{
	string str = sc.yytext;
	$$ = new PrefixedNameNode() { Pname_Ns = str.Split(':')[0], Pn_Local = str.Substring(str.IndexOf(':') + 1) };
}
| PNAME_NS
{
	$$ = new PrefixedNameNode() { Pname_Ns = sc.yytext.Split(':')[0] };
}
;

/* [69] */
BlankNode
: BLANK_NODE_LABEL
{
	$$ = new BlankNode(){ Value = sc.yytext };
}
| ANON
{
	$$ = new BlankNode();
}
;

%%