%using SharQL.Utils;

/* Defines the generated scanner namespace. */
%namespace SharQL

/* Enables unicode characters. */
%option unicode

%{

	// Called each time a token has been returned.
	internal void LoadYylval()
	{
		// Collects token location information and makes it available
		// for the parserthrough the yylloc variable.
		yylloc = new CustomLexLocation(yyline, yycol, yyline, yycol + yyleng);
	}

	// Called by the parser on the occurence of an unexpected or missing 
	// token.
	public override void yyerror(string s, params object[] a)
	{
		// Adds the reported error to the error handler.
		handler.AddError(s, new MPLEX.Parser.LexSpan(tokLin, tokCol, tokLin, 
			tokECol, tokPos, tokEPos, new ScanBuffProxy(buffer)));
	}
	
	// Proxy class for the MPLEX.Lexer.ScanBuff class to allow for the 
	// MPLEX.Parser.LexSpan class to be used with the generated 
	// SharQL.ScanBuff class.
	public class ScanBuffProxy : MPLEX.Lexer.ScanBuff
	{
		private ScanBuff buffer;
	
		public override int Pos
		{
			get { return buffer.Pos; }
			set { buffer.Pos = value; }
		}
		
		public override int Read()
		{
			return buffer.Read();
		}
		
		public override int Peek()
		{
			return buffer.Peek();
		}
		
		public override int ReadPos
		{
			get { return buffer.ReadPos; }
		}
		
		public override string GetString(int b, int e)
		{
			return buffer.GetString(b, e);
		}
		
		public ScanBuffProxy(ScanBuff buffer)
		{
			this.buffer = buffer;
		}
	}

%}

/*
 * SPARQL Query Language for RDF
 * W3C Recommendation 15 January 2008
 * 
 * http://www.w3.org/TR/2008/REC-rdf-sparql-query-20080115/ 
 *
 */

iri_ref					<([^<>"{}|^`\\\u0000-\u0020])*>
langtag					@[a-zA-Z]+(\-[a-zA-Z0-9]+)*
exponent				[eE][\+\-]?[0-9]+
integer					[0-9]+
integer_positive		\+{integer}
integer_negative		\-{integer}
decimal					[0-9]+\.[0-9]*|\.[0-9]+
decimal_positive		\+{decimal}
decimal_negative		\-{decimal}
double					[0-9]+\.[0-9]*{exponent}|\.([0-9])+{exponent}|([0-9])+{exponent}
double_positive			\+{double}
double_negative			\-{double}
echar					\\[tbnrf\\"']

/* Range 10000-EFFFF is left out, as MPLEX currently only supports 16-bit unicode characters. */
pn_chars_base			[A-Z]|[a-z]|[\u00C0-\u00D6]|[\u00D8-\u00F6]|[\u00F8-\u02FF]|[\u0370-\u037D]|[\u037F-\u1FFF]|[\u200C-\u200D]|[\u2070-\u218F]|[\u2C00-\u2FEF]|[\u3001-\uD7FF]|[\uF900-\uFDCF]|[\uFDF0-\uFFFD]

pn_chars_u				{pn_chars_base}|_
pn_chars				{pn_chars_u}|\-|[0-9]|\u00B7|[\u0300-\u036F]|[\u203F-\u2040]
pn_prefix				{pn_chars_base}(({pn_chars}|\.)*{pn_chars})?
pn_local				({pn_chars_u}|[0-9])(({pn_chars}|\.)*{pn_chars})?
pname_ns				{pn_prefix}?:
pname_ln				{pname_ns}{pn_local}
blank_node_label		_:{pn_local}
varname					({pn_chars_u}|[0-9])({pn_chars_u}|[0-9]|\u00B7|[\u0300-\u036F]|[\u203F-\u2040])*
var1					\?{varname}
var2					\${varname}
string_literal1			'(([^\u0027\u005C\u000A\u000D])|{echar})*'
string_literal2 		\"(([^\u0022\u005C\u000A\u000D])|{echar})*\"
string_literal_long1	'''(('|'')?([^'\\]|{echar}))*'''
string_literal_long2	\"\"\"((\"|\"\")?([^\"\\]|{echar}))*\"\"\"
ws						\u0020|\u0009|\u000D|\u000A
nil						\({ws}*\)
anon					\[{ws}*\]
eq						"="
less					"<"
greater					">"
or						"||"
and						"&&"
not						"!="
lessorequal				"<="
greaterorequal			">="
hats					"^^"
comment					#[^\n]*

/*
 * A.8 Grammar
 *
 * (...) Keywords are matched in a case-insensitive manner with the 
 * exception of the keyword 'a'.
 */
base					[Bb][Aa][Ss][Ee]
prefix					[Pp][Rr][Ee][Ff][Ii][Xx]
select					[Ss][Ee][Ll][Ee][Cc][Tt]
distinct				[Dd][Ii][Ss][Tt][Ii][Nn][Cc][Tt]
reduced					[Rr][Ee][Dd][Uu][Cc][Ee][Dd]
construct				[Cc][Oo][Nn][Ss][Tt][Rr][Uu][Cc][Tt]
describe				[Dd][Ee][Ss][Cc][Rr][Ii][Bb][Ee]
ask						[Aa][Ss][Kk]
from					[Ff][Rr][Oo][Mm]
named					[Nn][Aa][Mm][Ee][Dd]
where					[Ww][Hh][Ee][Rr][Ee]
orderby					[Oo][Rr][Dd][Ee][Rr]\u0020[Bb][Yy]
asc						[Aa][Ss][Cc]
desc					[Dd][Ee][Ss][Cc]
limit					[Ll][Ii][Mm][Ii][Tt]
offset					[Oo][Ff][Ff][Ss][Ee][Tt]
optional				[Oo][Pp][Tt][Ii][Oo][Nn][Aa][Ll]
graph					[Gg][Rr][Aa][Pp][Hh]
filter					[Ff][Ii][Ll][Tt][Ee][Rr]
union					[Uu][Nn][Ii][Oo][Nn]
str						[Ss][Tt][Rr]
lang					[Ll][Aa][Nn][Gg]
langmatches				[Ll][Aa][Nn][Gg][Mm][Aa][Tt][Cc][Hh][Ee][Ss]
datatype				[Dd][Aa][Tt][Aa][Tt][Yy][Pp][Ee]
bound					[Bb][Oo][Uu][Nn][Dd]
sameterm				[Ss][Aa][Mm][Ee][Tt][Ee][Rr][Mm]
isiri					[Ii][Ss][Ii][Rr][Ii]
isuri					[Ii][Ss][Uu][Rr][Ii]
isblank					[Ii][Ss][Bb][Ll][Aa][Nn][Kk]
isliteral				[Ii][Ss][Ll][Ii][Tt][Ee][Rr][Aa][Ll]
regex					[Rr][Ee][Gg][Ee][Xx]
true					[Tt][Rr][Uu][Ee]
false					[Ff][Aa][Ll][Ss][Ee]

%%

{base}					{return (int)Tokens.BASE;}		
{prefix}				{return (int)Tokens.PREFIX;}
{select}				{return (int)Tokens.SELECT;}
{distinct}				{return (int)Tokens.DISTINCT;}
{reduced}				{return (int)Tokens.REDUCED;}
{construct}				{return (int)Tokens.CONSTRUCT;}
{describe}				{return (int)Tokens.DESCRIBE;}
{ask}					{return (int)Tokens.ASK;}
{from}					{return (int)Tokens.FROM;}
{named}					{return (int)Tokens.NAMED;}
{where}					{return (int)Tokens.WHERE;}
{orderby}				{return (int)Tokens.ORDERBY;}
{asc}					{return (int)Tokens.ASC;}
{desc}					{return (int)Tokens.DESC;}
{limit}					{return (int)Tokens.LIMIT;}
{offset}				{return (int)Tokens.OFFSET;}
{optional}				{return (int)Tokens.OPTIONAL;}
{graph}					{return (int)Tokens.GRAPH;}
{filter}				{return (int)Tokens.FILTER;}
{union}					{return (int)Tokens.UNION;}
{eq}					{return (int)'=';}
{less}					{return (int)'<';}
{greater}				{return (int)'>';}					
{or}					{return (int)Tokens.OR;}
{and}					{return (int)Tokens.AND;}
{not}					{return (int)Tokens.NOT;}
{lessorequal}			{return (int)Tokens.LESSOREQUAL;}
{greaterorequal}		{return (int)Tokens.GREATEROREQUAL;}
{str}					{return (int)Tokens.STR;}
{lang}					{return (int)Tokens.LANG;}
{langmatches}			{return (int)Tokens.LANGMATCHES;}
{datatype}				{return (int)Tokens.DATATYPE;}
{bound}					{return (int)Tokens.BOUND;}
{sameterm}				{return (int)Tokens.SAMETERM;}
{isiri}					{return (int)Tokens.ISIRI;}
{isuri}					{return (int)Tokens.ISURI;}
{isblank}				{return (int)Tokens.ISBLANK;}
{isliteral}				{return (int)Tokens.ISLITERAL;}
{regex}					{return (int)Tokens.REGEX;}
{true}					{return (int)Tokens.TRUE;}
{false}					{return (int)Tokens.FALSE;}
{hats}					{return (int)Tokens.HATS;}
{iri_ref}				{return (int)Tokens.IRI_REF;}
{langtag}				{return (int)Tokens.LANGTAG;}
{exponent}				{return (int)Tokens.EXPONENT;}
{integer}				{return (int)Tokens.INTEGER;}
{integer_positive}		{return (int)Tokens.INTEGER_POSITIVE;}
{integer_negative}		{return (int)Tokens.INTEGER_NEGATIVE;}
{decimal}				{return (int)Tokens.DECIMAL;}
{decimal_positive}		{return (int)Tokens.DECIMAL_POSITIVE;}
{decimal_negative}		{return (int)Tokens.DECIMAL_NEGATIVE;}
{double}				{return (int)Tokens.DOUBLE;}
{double_positive}		{return (int)Tokens.DOUBLE_POSITIVE;}
{double_negative}		{return (int)Tokens.DOUBLE_NEGATIVE;}
{echar}					{return (int)Tokens.ECHAR;}
{pname_ns}				{return (int)Tokens.PNAME_NS;}
{pname_ln}				{return (int)Tokens.PNAME_LN;}
{blank_node_label}		{return (int)Tokens.BLANK_NODE_LABEL;}
{var1}					{return (int)Tokens.VAR1;}
{var2}					{return (int)Tokens.VAR2;}
{string_literal1}		{return (int)Tokens.STRING_LITERAL1;}
{string_literal2}		{return (int)Tokens.STRING_LITERAL2;}
{string_literal_long1}	{return (int)Tokens.STRING_LITERAL_LONG1;}
{string_literal_long2}	{return (int)Tokens.STRING_LITERAL_LONG2;}
\*						{return (int)'*';}
\{						{return (int)'{';}
\}						{return (int)'}';}	
\(						{return (int)'(';}
\)						{return (int)')';}	
,						{return (int)',';}
;						{return (int)';';}
\.						{return (int)'.';}
a						{return (int)'a';}
\[						{return (int)'[';}
\]						{return (int)']';}
\+						{return (int)'+';}
\-						{return (int)'-';}
\!						{return (int)'!';}
\/						{return (int)'/';}
{ws}					{/*IGNORE*/}
{nil}					{return (int)Tokens.NIL;}
{anon}					{return (int)Tokens.ANON;}
{comment}				{/*IGNORE*/}

/*
 * Match for any single character not matching any other token definition.
 * (Makes parser aware of undefined tokens.)
 */
.						{return (int)yytext[0];}

/* Special-case match for end of file. */
<<EOF>>					{return (int)Tokens.EOF;}

%{
	// Called each time a token has been returned.
	LoadYylval();
%}

%%