﻿using System;
using System.Collections.Generic;
using System.Text;

namespace SharQL.Utils
{
	/// <summary>
	/// Contains methods for resolving escape sequences in SPARQL queries.
	/// </summary>
	internal static class EscapeSequenceResolver
	{
		/// <summary>Recognized escape sequences.</summary>
		private static Dictionary<string, char> escapes = new Dictionary<string, char>()
		{
			{@"\t", '\u0009'},
			{@"\n", '\u000A'},
			{@"\r", '\u000D'},
			{@"\b", '\u0008'},
			{@"\""", '\u0022'},
			{@"\'", '\u0027'},
			{@"\\", '\u005C'}
		};

		/// <summary>
		/// Resolves common escape sequences according to appendix A.7 in the
		/// SPARQL specification (http://www.w3.org/TR/rdf-sparql-query/).
		/// </summary>
		/// <param name="escapedString">The string to resolve</param>
		/// <returns>The resolved string</returns>
		/// <exception cref="ArgumentException">If specified string contains
		/// unrecognized escape sequences.</exception>
		internal static string ResolveCommonEscapeSequences(string escapedString)
		{
			StringBuilder sb = new StringBuilder();

			for (int i = 0; i < escapedString.Length; i++)
			{
				try
				{
					if (escapedString[i] != '\\')
					{
						sb.Append(escapedString[i]);
					}
					else
					{
						string escape = string.Empty + escapedString[i++] + escapedString[i];
						if (!escapes.ContainsKey(escape))
							throw new ArgumentException("Input contains unrecognized escape sequence: " + escape);
						sb.Append(escapes[escape]);
					}
				}
				catch (IndexOutOfRangeException)
				{
					throw new ArgumentException("Input string terminates when expectig escape sequence");
				}
			}
			return sb.ToString();
		}

		/// <summary>
		/// Resolves codepoint escape sequences according to appendix A.2 in the
		/// SPARQL specification (http://www.w3.org/TR/rdf-sparql-query/).
		/// </summary>
		/// <param name="escapedString">The string to resolve</param>
		/// <returns>The resolved string</returns>
		/// <exception cref="ArgumentException">If specified string contains
		/// illegal codepoint escape sequences.</exception>
		/// <remarks>Unlike the <c>ResolveRegularEscapeSequences</c> method, 
		/// unrecocgnized two-character escape sequences are ignores by this 
		/// method.</remarks>
		internal static string ResolveCodepointEscapeSequences(string escapedString)
		{
			StringBuilder sb = new StringBuilder();

			for (int i = 0; i < escapedString.Length; i++)
			{
				try
				{
					if (escapedString[i] != '\\')
					{
						sb.Append(escapedString[i]);
					}
					else
					{
						if (escapedString[i + 1] == 'u')
						{
							char unicodeChar;

							try
							{
								string codepointString = string.Empty + escapedString[i + 2] + escapedString[i + 3] + escapedString[i + 4] + escapedString[i + 5];
								int codepoint = Convert.ToInt32(codepointString, 16);
								unicodeChar = Convert.ToChar(codepoint);
							}
							catch (Exception)
							{
								throw new ArgumentException("Input contains illegal codepoint escape sequence");
							}

							sb.Append(unicodeChar);
							i += 5;
						}
						else if (escapedString[i + 1] == 'U')
						{
							char unicodeChar;

							try
							{
								string codepointString = string.Empty + escapedString[i + 2] + escapedString[i + 3] + escapedString[i + 4] + escapedString[i + 5] + escapedString[i + 6] + escapedString[i + 7] + escapedString[i + 8] + escapedString[i + 9];
								int codepoint = Convert.ToInt32(codepointString, 16);
								unicodeChar = Convert.ToChar(codepoint);
							}
							catch (Exception)
							{
								throw new ArgumentException("Input contains illegal codepoint escape sequence");
							}

							sb.Append(unicodeChar);
							i += 9;
						}
						else
						{
							sb.Append(escapedString[i++]);
							sb.Append(escapedString[i]);
						}
					}
				}
				catch (IndexOutOfRangeException)
				{
					throw new ArgumentException("Input string terminates when expectig escape sequence");
				}
			}
			return sb.ToString();
		}
	}
}
