﻿using System.Collections.Generic;
using System.IO;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SharQL;

namespace SharQLTest
{
	/// <summary>
	/// Summary description for ScannerTest
	/// </summary>
	[TestClass]
	public class ScannerTest
	{
		private TestContext testContextInstance;
		private Scanner scanner;

		public ScannerTest()
		{
			scanner = new Scanner();
		}

		/// <summary>
		/// Gets or sets the test context which provides
		/// information about and functionality for the current test run.
		/// </summary>
		public TestContext TestContext
		{
			get
			{
				return testContextInstance;
			}
			set
			{
				testContextInstance = value;
			}
		}

		#region Additional test attributes
		//
		// You can use the following additional attributes as you write your tests:
		//
		// Use ClassInitialize to run code before running the first test in the class
		// [ClassInitialize()]
		// public static void MyClassInitialize(TestContext testContext) { }
		//
		// Use ClassCleanup to run code after all tests in a class have run
		// [ClassCleanup()]
		// public static void MyClassCleanup() { }
		//
		// Use TestInitialize to run code before running each test 
		// [TestInitialize()]
		// public void MyTestInitialize() { }
		//
		// Use TestCleanup to run code after each test has run
		// [TestCleanup()]
		// public void MyTestCleanup() { }
		//
		#endregion

		private Tokens[] scanInput(string input)
		{
			List<Tokens> tokens = new List<Tokens>();

			scanner.SetSource(input, 0);
			int token = scanner.yylex();
			while (token != (int)Tokens.EOF)
			{
				tokens.Add((Tokens)token);
				token = scanner.yylex();
			}
			return tokens.ToArray();
		}

		private bool areEqual(Tokens[] expected, Tokens[] actual)
		{
			if (expected.Length != actual.Length)
				return false;

			for (int i = 0; i < expected.Length; i++)
				if (expected[i] != actual[i])
					return false;

			return true;
		}

		private bool assertScannerOutput(string inputFile, Tokens[] expectedOutput)
		{
			string input = File.ReadAllText(inputFile);
			Tokens[] actualOutput = scanInput(input);
			return areEqual(expectedOutput, actualOutput);
		}

		[TestMethod]
		public void Scanner_1_basic_01()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-basic-01.rq",
				new Tokens[] 
				{ 
					Tokens.SELECT, 
					(Tokens)((int)'*'), 
					Tokens.WHERE, 
					(Tokens)((int)'{'), 
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_basic_06()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-basic-06.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_bnodes_03()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-bnodes-03.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'['),
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)']'),
					Tokens.PNAME_LN,
					(Tokens)((int)'['),
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)']'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_expr_04()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-expr-04.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					Tokens.FILTER,
					Tokens.PNAME_LN,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_forms_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-forms-02.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'('),
					Tokens.ANON,
					Tokens.ANON,
					(Tokens)((int)')'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_lists_04()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-lists-04.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'('),
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')'),
					(Tokens)((int)')'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_lit_16()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-lit-16.rq",
				new Tokens[]
				{
					Tokens.BASE,
					Tokens.IRI_REF,
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.STRING_LITERAL_LONG1,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_order_06()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-order-06.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					Tokens.ORDERBY,
					Tokens.DESC,
					(Tokens)((int)'('),
					Tokens.VAR1,
					Tokens.INTEGER_POSITIVE,
					(Tokens)((int)')'),
					Tokens.PNAME_LN,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')'),
					Tokens.ASC,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')')
				}));
		}

		[TestMethod]
		public void Scanner_1_pat_03()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-pat-03.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.OPTIONAL,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					Tokens.UNION,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_qname_08()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-qname-08.rq",
				new Tokens[]
				{
					Tokens.BASE,
					Tokens.IRI_REF,
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_NS,
					Tokens.PNAME_NS,
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_struct_08()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-struct-08.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					(Tokens)((int)'{'),
					Tokens.OPTIONAL,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_1_struct_14()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-struct-14.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.OPTIONAL,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.OPTIONAL,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)'}'),
					(Tokens)((int)'}'),
				}));
		}

		[TestMethod]
		public void Scanner_1_union_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql1\syntax-union-02.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					(Tokens)((int)'{'),
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					Tokens.UNION,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					Tokens.UNION,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_bnode_03()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-bnode-03.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'['),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)']'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_dataset_04()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-dataset-04.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					Tokens.VAR1,
					Tokens.FROM,
					Tokens.PNAME_LN,
					Tokens.FROM,
					Tokens.PNAME_LN,
					Tokens.FROM,
					Tokens.NAMED,
					Tokens.PNAME_LN,
					Tokens.FROM,
					Tokens.NAMED,
					Tokens.PNAME_LN,
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_form_construct_01()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-form-construct01.rq",
				new Tokens[]
				{
					Tokens.CONSTRUCT,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.IRI_REF,
					Tokens.IRI_REF,
					(Tokens)((int)'.'),
					Tokens.VAR1,
					Tokens.IRI_REF,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_form_describe_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-form-describe02.rq",
				new Tokens[]
				{
					Tokens.DESCRIBE,
					Tokens.IRI_REF,
					Tokens.VAR1,
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.IRI_REF,
					Tokens.IRI_REF,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_function_04()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-function-04.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.PNAME_LN,
					(Tokens)((int)'('),
					Tokens.INTEGER,
					(Tokens)((int)')'),
					(Tokens)((int)')'),
					(Tokens)((int)'.'),
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.PNAME_LN,
					(Tokens)((int)'('),
					Tokens.INTEGER,
					(Tokens)((int)','),
					Tokens.INTEGER,
					(Tokens)((int)')'),
					(Tokens)((int)')'),
					(Tokens)((int)'.'),
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.PNAME_LN,
					(Tokens)((int)'('),
					Tokens.INTEGER,
					(Tokens)((int)','),
					Tokens.INTEGER,
					(Tokens)((int)')'),
					(Tokens)((int)')'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_general_11()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-general-11.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.IRI_REF,
					Tokens.IRI_REF,
					Tokens.DOUBLE_NEGATIVE,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_general_14()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-general-14.rq",
				new Tokens[]
				{
					Tokens.BASE,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.IRI_REF,
					Tokens.IRI_REF,
					Tokens.IRI_REF,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_graph_05()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-graph-05.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.GRAPH,
					Tokens.VAR1,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					Tokens.GRAPH,
					Tokens.VAR1,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.VAR1,
					(Tokens)((int)'}'),
					(Tokens)((int)'}'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_keywords_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-keywords-02.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.PNAME_LN,
					Tokens.VAR1,
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_2_lists_04()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql2\syntax-lists-04.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'('),
					Tokens.INTEGER,
					Tokens.INTEGER,
					(Tokens)((int)')'),
					Tokens.PNAME_LN,
					Tokens.INTEGER,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_3_syn_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql3\syn-02.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_3_syn_05()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql3\syn-05.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)'>'),
					Tokens.INTEGER,
					(Tokens)((int)')'),
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_3_syn_07()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql3\syn-07.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					Tokens.PNAME_LN,
					(Tokens)((int)';'),
					Tokens.FILTER,
					(Tokens)((int)'('),
					Tokens.VAR1,
					(Tokens)((int)')'),
					(Tokens)((int)'}')
			}));
		}

		[TestMethod]
		public void Scanner_4_syn_10()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql4\syn-10.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					(Tokens)((int)'*'),
					Tokens.WHERE,
					(Tokens)((int)'{'),
					(Tokens)((int)'{'),
					Tokens.BLANK_NODE_LABEL,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					Tokens.BLANK_NODE_LABEL,
					Tokens.VAR1,
					Tokens.BLANK_NODE_LABEL,
					(Tokens)((int)'}'),
					Tokens.UNION,
					(Tokens)((int)'{'),
					Tokens.BLANK_NODE_LABEL,
					Tokens.VAR1,
					Tokens.BLANK_NODE_LABEL,
					(Tokens)((int)'}'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_4_syn_leading_digits_in_prefixed_names()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql4\syn-leading-digits-in-prefixed-names.rq",
				new Tokens[]
				{
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.PREFIX,
					Tokens.PNAME_NS,
					Tokens.IRI_REF,
					Tokens.SELECT,
					Tokens.VAR1,
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.PNAME_LN,
					(Tokens)((int)'a'),
					Tokens.PNAME_LN,
					(Tokens)((int)';'),
					Tokens.PNAME_LN,
					Tokens.VAR1,
					(Tokens)((int)'.'),
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_5_reduced_02()
		{
			Assert.IsTrue(assertScannerOutput(@"..\..\..\Tests\syntax-sparql5\syntax-reduced-02.rq",
				new Tokens[]
				{
					Tokens.SELECT,
					Tokens.REDUCED,
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.WHERE,
					(Tokens)((int)'{'),
					Tokens.VAR1,
					Tokens.VAR1,
					Tokens.VAR1,
					(Tokens)((int)'}')
				}));
		}

		[TestMethod]
		public void Scanner_keywords()
		{
			string keywords = @"BASE SELECT ORDER BY FROM GRAPH STR isURI
								PREFIX CONSTRUCT LIMIT FROM NAMED OPTIONAL
								LANG isIRI DESCRIBE OFFSET WHERE UNION 
								LANGMATCHES isLITERAL ASK DISTINCT FILTER
								DATATYPE REGEX REDUCED a BOUND sameTERM 
								true false";
			Tokens[] actual = scanInput(keywords);
			Tokens[] expected = new Tokens[]
			{
				Tokens.BASE,
				Tokens.SELECT,
				Tokens.ORDERBY,
				Tokens.FROM,
				Tokens.GRAPH,
				Tokens.STR,
				Tokens.ISURI,
				Tokens.PREFIX,
				Tokens.CONSTRUCT,
				Tokens.LIMIT,
				Tokens.FROM,
				Tokens.NAMED,
				Tokens.OPTIONAL,
				Tokens.LANG,
				Tokens.ISIRI,
				Tokens.DESCRIBE,
				Tokens.OFFSET,
				Tokens.WHERE,
				Tokens.UNION,
				Tokens.LANGMATCHES,
				Tokens.ISLITERAL,
				Tokens.ASK,
				Tokens.DISTINCT,
				Tokens.FILTER,
				Tokens.DATATYPE,
				Tokens.REGEX,
				Tokens.REDUCED,
				(Tokens)((int)'a'),
				Tokens.BOUND,
				Tokens.SAMETERM,
				Tokens.TRUE,
				Tokens.FALSE
			};
			Assert.IsTrue(areEqual(expected, actual));
		}

		[TestMethod]
		public void Scanner_comments()
		{
			string comment = "# comments start with a hash sign and lasts for entire line\n?var";
			Tokens[] actual = scanInput(comment);
			Tokens[] expected = new Tokens[] 
			{
				Tokens.VAR1
			};
			Assert.IsTrue(areEqual(expected, actual));

			comment = "?var # comment for variable should be ignored";
			actual = scanInput(comment);
			expected = new Tokens[] 
			{
				Tokens.VAR1
			};
			Assert.IsTrue(areEqual(expected, actual));

			comment = "<iri#> # hash sign inside iri_ref should not be treated as comment";
			actual = scanInput(comment);
			expected = new Tokens[]
			{
				Tokens.IRI_REF
			};
			Assert.IsTrue(areEqual(expected, actual));

			comment = "'''Long string literal with a # sign''' # hash sign inside string literals should not be treated as comment";
			actual = scanInput(comment);
			expected = new Tokens[]
			{
				Tokens.STRING_LITERAL_LONG1
			};
			Assert.IsTrue(areEqual(expected, actual));
		}
	}
}
