﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using SharQL;
using SharQL.Ast;
using SharQL.Ast.Nodes;
using SharQL.Ast.Visitor;

namespace SharQLTestClient
{
	public partial class SharQLTestClient : Form
	{
		private SharQL.Parser parser;

		private Func<string, bool> parseDelegate;
		private Action<IVisitor> visitorDelegate;

		public SharQLTestClient()
		{
			InitializeComponent();
			parser = new SharQL.Parser();
		}

		#region Event Handlers

		private void SharQLTestClient_Load(object sender, EventArgs e)
		{
			Assembly sharqlAssembly = Assembly.GetAssembly(typeof(SharQL.Parser));
			Assembly mppgAssembly = Assembly.GetAssembly(typeof(mpcc.Parser));
			Assembly mplexAssembly = Assembly.GetAssembly(typeof(MPLEX.Lexer.Scanner));

			toolStripStatusLabelSharqlVersion.Text = string.Format(toolStripStatusLabelSharqlVersion.Text, sharqlAssembly.GetName().Version.ToString(2));
			toolStripStatusLabelMppgVersion.Text = string.Format(toolStripStatusLabelMppgVersion.Text, mppgAssembly.GetName().Version.ToString(2));
			toolStripStatusLabelMplexVersion.Text = string.Format(toolStripStatusLabelMplexVersion.Text, mplexAssembly.GetName().Version.ToString(2));
		}

		private void buttonBrowse_Click(object sender, EventArgs e)
		{
			if (DialogResult.OK == openFileDialog.ShowDialog(this))
			{
				textBoxSparqlFile.Text = openFileDialog.FileName;
			}
		}

		private void buttonParse_Click(object sender, EventArgs e)
		{
			parseInputFile();
		}

		private void buttonScan_Click(object sender, EventArgs e)
		{
			scanInputFile();
		}

		private void expandAllToolStripMenuItem_Click(object sender, EventArgs e)
		{
			treeViewAST.ExpandAll();
		}

		private void collapseAllToolStripMenuItem_Click(object sender, EventArgs e)
		{
			treeViewAST.CollapseAll();
		}

		private void contextMenuStrip_Opening(object sender, System.ComponentModel.CancelEventArgs e)
		{
			collapseAllToolStripMenuItem.Enabled = expandAllToolStripMenuItem.Enabled = treeViewAST.Nodes.Count > 0;
		}

		private void treeViewAST_AfterSelect(object sender, TreeViewEventArgs e)
		{
			NodeBase node = e.Node.Tag as NodeBase;
			if (node != null)
			{
				textBoxDetails.Text = getDetails(node);
			}
			else
			{
				textBoxDetails.Text = string.Empty;
			}
		}

		#endregion

		#region Parse Control Methods

		private void parseInputFile()
		{
			disableControls();

			statusStripStatusLabel.Text = "Reading input file...";
			Refresh();

            string source;

			try
			{
				using (FileStream fs = new FileStream(textBoxSparqlFile.Text, FileMode.Open))
				using (StreamReader reader = new StreamReader(fs))
				{
                    source = reader.ReadToEnd();
				}
				parseDelegate = parser.Parse;

				statusStripStatusLabel.Text = "Parsing...";
				parseDelegate.BeginInvoke(source, parseCompleted, null);
			}
			catch (Exception e)
			{
				MessageBox.Show(this, e.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);

				statusStripStatusLabel.Text = "Ready";
				enableControls();
			}
		}

		private void parseCompleted(IAsyncResult ar)
		{
			if (InvokeRequired)
			{
				Invoke(new AsyncCallback(parseCompleted), ar);
			}
			else
			{
				try
				{
					bool result = parseDelegate.EndInvoke(ar);
					if (result && parser.AstRoot != null)
					{
						TreeBuilderVisitor tbv = new TreeBuilderVisitor();
						visitorDelegate = parser.AstRoot.Accept;

						statusStripStatusLabel.Text = "Visualizing...";
						visitorDelegate.BeginInvoke(tbv, treeCompleted, tbv);
					}
					else
					{
						MessageBox.Show(this, "Unable to parse input file", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
						showErrors(parser.Errors);

						statusStripStatusLabel.Text = "Ready";
						enableControls();
					}
				}
				catch (Exception e)
				{
					MessageBox.Show(this, e.Message, "Exception!", MessageBoxButtons.OK, MessageBoxIcon.Error);

					statusStripStatusLabel.Text = "Ready";
					enableControls();
				}
			}
		}

		private void treeCompleted(IAsyncResult ar)
		{
			if (InvokeRequired)
			{
				Invoke(new AsyncCallback(treeCompleted), ar);
			}
			else
			{
				TreeBuilderVisitor tbv = ar.AsyncState as TreeBuilderVisitor;
				treeViewAST.Nodes.Clear();
				treeViewAST.Nodes.Add(tbv.TreeNodes[parser.AstRoot]);

				statusStripStatusLabel.Text = "Ready";
				enableControls();
			}
		}

		#endregion

		#region Private Helper Methods

		private void scanInputFile()
		{
			Scanner scanner = new Scanner();
			List<Tokens> tokenList = new List<Tokens>();
			string input = File.ReadAllText(textBoxSparqlFile.Text);
			
			scanner.SetSource(input, 0);
			int token = scanner.yylex();
			while (token != (int)Tokens.EOF)
			{
				tokenList.Add((Tokens)token);
				token = scanner.yylex();
			}

			ScannerResult res = new ScannerResult();
			res.SetTokens(tokenList.ToArray());
			res.Show();
		}

		private void disableControls()
		{
			textBoxSparqlFile.Enabled = false;
			buttonBrowse.Enabled = false;
			buttonParse.Enabled = false;
			splitContainerRoot.Enabled = false;
		}

		private void enableControls()
		{
			textBoxSparqlFile.Enabled = true;
			buttonBrowse.Enabled = true;
			buttonParse.Enabled = true;
			splitContainerRoot.Enabled = true;
		}

		private void showErrors(ReadOnlyCollection<MPLEX.Parser.Error> errors)
		{
			listViewErrors.SuspendLayout();
			listViewErrors.Items.Clear();
			if (errors != null)
			{
				foreach (MPLEX.Parser.Error err in errors)
				{
					listViewErrors.Items.Add(new ListViewItem(new string[] { err.message, err.span.sLin.ToString(), err.span.sCol.ToString(), err.span.sPos.ToString() }));
				}
			}
			listViewErrors.ResumeLayout();
		}

		private string getDetails(NodeBase node)
		{
			StringBuilder sb = new StringBuilder();

			sb.AppendLine("Node Type:");
			sb.AppendLine(node.ToString());
			sb.AppendLine();

			if (node is IriRefNode)
			{
				sb.AppendLine("IRI Reference:");
				sb.AppendLine(((IriRefNode)node).Value);
			}
			else if (node is VarNode)
			{
				sb.AppendLine("Variable Name:");
				sb.AppendLine(((VarNode)node).Value);
			}
			else if (node is StarVarListNode)
			{
				sb.AppendLine("Is Star:");
				sb.AppendLine(((StarVarListNode)node).IsStar.ToString());
			}
			else if (node is StarVarOrIRIrefListNode)
			{
				sb.AppendLine("Is Star:");
				sb.AppendLine(((StarVarOrIRIrefListNode)node).IsStar.ToString());
			}
			else if (node is PrefixDeclNode)
			{
				sb.AppendLine("Namespace:");
				sb.AppendLine(((PrefixDeclNode)node).Namespace);
				sb.AppendLine();
				sb.AppendLine("IsBaseDecl:");
				sb.AppendLine(((PrefixDeclNode)node).IsBaseDecl.ToString());
			}
			else if (node is PrefixedNameNode)
			{
				sb.AppendLine("Namespace:");
				sb.AppendLine(((PrefixedNameNode)node).Pname_Ns);
				sb.AppendLine();
				sb.AppendLine("Local Name:");
				sb.AppendLine(((PrefixedNameNode)node).Pn_Local);
			}
			else if (node is BlankNode)
			{
				sb.AppendLine("Value");
				sb.AppendLine(((BlankNode)node).Value);
			}
			else if (node is AscOrDescNode)
			{
				sb.AppendLine("Sort Order:");
				sb.AppendLine(((AscOrDescNode)node).Order.ToString());
			}
			else if (node is BooleanLiteralNode)
			{
				sb.AppendLine("Literal:");
				sb.AppendLine(((BooleanLiteralNode)node).Value.ToString());
			}
			else if (node is NumericLiteralNode)
			{
				sb.AppendLine("Literal:");
				NumericLiteralNode nln = (NumericLiteralNode)node;
				switch (nln.Type)
				{
					case NumericLiteralNode.LiteralType.Int:
						sb.AppendLine(nln.IntValue.ToString());
						break;
					case NumericLiteralNode.LiteralType.Double:
						sb.AppendLine(nln.DoubleValue.ToString());
						break;
					case NumericLiteralNode.LiteralType.Decimal:
						sb.AppendLine(nln.DecimalValue.ToString());
						break;
					default:
						sb.AppendLine("Error....");
						break;
				}
			}
			else if (node is RdfLiteralNode)
			{
				sb.AppendLine("Language Tag:");
				if (string.IsNullOrEmpty(((RdfLiteralNode)node).LangTag))
					sb.AppendLine("[none]");
				else
					sb.AppendLine(((RdfLiteralNode)node).LangTag);
			}
			else if (node is StringLiteralNode)
			{
				sb.AppendLine("String:");
				sb.AppendLine(((StringLiteralNode)node).Value);
			}
			else if (node is BuiltInCallNode)
			{
				sb.AppendLine("Function:");
				sb.AppendLine(((BuiltInCallNode)node).Type.ToString());
			}
			else if (node is DefaultOrNamedGraphClauseNode)
			{
				sb.AppendLine("Type:");
				sb.AppendLine(((DefaultOrNamedGraphClauseNode)node).Type.ToString());
			}
			else if (node is DistinctReducedModifierNode)
			{
				sb.AppendLine("Type:");
				sb.AppendLine(((DistinctReducedModifierNode)node).Modifier.ToString());
			}
			else if (node is LimitClauseNode)
			{
				sb.AppendLine("Limit:");
				sb.AppendLine(((LimitClauseNode)node).Limit.ToString());
			}
			else if (node is OffsetClauseNode)
			{
				sb.AppendLine("Offset:");
				sb.AppendLine(((OffsetClauseNode)node).Offset.ToString());
			}
			else if (node is UnaryExpressionNode)
			{
				sb.AppendLine("Operator:");
				sb.AppendLine(((UnaryExpressionNode)node).Operator.ToString());
			}
			else if (node is UnaryExpressionListNode)
			{
				sb.AppendLine("Operator:");
				sb.AppendLine(((UnaryExpressionListNode)node).Operator.ToString());
			}
			else if (node is RelationalExpressionNode)
			{
				sb.AppendLine("Operator:");
				sb.AppendLine(((RelationalExpressionNode)node).Operator.ToString());
			}
			else if (node is AdditiveExpressionListNode)
			{
				sb.AppendLine("Operator:");
				sb.AppendLine(((AdditiveExpressionListNode)node).Operator.ToString());
			}

			return sb.ToString();
		}

		#endregion
	}
}
